import uharfbuzz as hb
from pathlib import Path
import sys
import platform
import pytest


TESTDATA = Path(__file__).parent / "data"
ADOBE_BLANK_TTF_PATH = TESTDATA / "AdobeBlank.subset.ttf"
OPEN_SANS_TTF_PATH = TESTDATA / "OpenSans.subset.ttf"
MUTATOR_SANS_TTF_PATH = TESTDATA / "MutatorSans-VF.subset.ttf"
SPARSE_FONT_TTF_PATH = TESTDATA / "SparseFont.ttf"
MATH_FONT_TTF_PATH = TESTDATA / "STIXTwoMath-Regular.ttf"
COLORv0_FONT_TTF_PATH = TESTDATA / "chromacheck-colr.ttf"
COLORv1_FONT_TTF_PATH = TESTDATA / "test_glyphs-glyf_colr_1.ttf"


@pytest.fixture
def blankfont():
    """Return a subset of AdobeBlank.ttf containing the following glyphs/characters:
    [
        {gid=0, name=".notdef"},
        {gid=1, name="a", code=0x61},
        {gid=2, name="b", code=0x62},
        {gid=3, name="c", code=0x63},
        {gid=4, name="d", code=0x64},
        {gid=5, name="e", code=0x65},
        {gid=6, name="ccedilla", code=0x62},
        {gid=7, name="uni0431", code=0x0431},  # CYRILLIC SMALL LETTER BE
        {gid=8, name="u1F4A9", code=0x1F4A9},  # PILE OF POO
    ]
    """
    blob = hb.Blob.from_file_path(ADOBE_BLANK_TTF_PATH)
    face = hb.Face(blob)
    font = hb.Font(face)
    return font


@pytest.fixture
def opensans():
    """Return a subset of OpenSans.ttf containing the following glyphs/characters:
    [
        {gid=0, name=".notdef"},
        {gid=1, name="A", code=0x41},
    ]
    """
    blob = hb.Blob(OPEN_SANS_TTF_PATH.read_bytes())
    face = hb.Face(blob)
    font = hb.Font(face)
    return font


@pytest.fixture
def mutatorsans():
    """Return a subset of MutatorSans-VF with a wdth and wght axis."""
    face = hb.Face(MUTATOR_SANS_TTF_PATH.read_bytes())
    font = hb.Font(face)
    return font


@pytest.fixture
def sparsefont():
    """Return a font that only has a few tables:
    GPOS, cmap, head, maxp and post"""
    face = hb.Face(SPARSE_FONT_TTF_PATH.read_bytes())
    font = hb.Font(face)
    return font


@pytest.fixture
def mathfont():
    """Return a subset of STIX Two Math font with only MATH, post, head, and maxp tables"""
    face = hb.Face(MATH_FONT_TTF_PATH.read_bytes())
    font = hb.Font(face)
    return font


@pytest.fixture
def colorv0font():
    blob = hb.Blob.from_file_path(COLORv0_FONT_TTF_PATH)
    face = hb.Face(blob)
    font = hb.Font(face)
    return font


@pytest.fixture
def colorv1font():
    blob = hb.Blob.from_file_path(COLORv1_FONT_TTF_PATH)
    face = hb.Face(blob)
    font = hb.Font(face)
    return font


class TestBuffer:
    def test_init(self):
        buf = hb.Buffer()

    def test_create(self):
        buf = hb.Buffer.create()

    @pytest.mark.parametrize(
        "string, expected",
        [
            ("abcde", [(0x61, 0), (0x62, 1), (0x63, 2), (0x64, 3), (0x65, 4)]),
            ("abçde", [(0x61, 0), (0x62, 1), (0xE7, 2), (0x64, 3), (0x65, 4)]),
            ("aбcde", [(0x61, 0), (0x431, 1), (0x63, 2), (0x64, 3), (0x65, 4)]),
            ("abc💩e", [(0x61, 0), (0x62, 1), (0x63, 2), (0x1F4A9, 3), (0x65, 4)]),
        ],
        ids=["ascii", "latin1", "ucs2", "ucs4"],
    )
    def test_add_str(self, string, expected):
        buf = hb.Buffer()
        buf.add_str(string)
        infos = [(g.codepoint, g.cluster) for g in buf.glyph_infos]
        flags = [(g.flags) for g in buf.glyph_infos]
        assert all(0 == f for f in flags)
        assert infos == expected

    def test_add_utf8(self):
        buf = hb.Buffer()
        buf.add_utf8("aбç💩e".encode("utf-8"))
        infos = [(g.codepoint, g.cluster) for g in buf.glyph_infos]
        assert infos == [(0x61, 0), (0x431, 1), (0xE7, 3), (0x1F4A9, 5), (0x65, 9)]

    def test_add_codepoints(self):
        buf = hb.Buffer()
        buf.add_codepoints([0x61, 0x431, 0xE7, 0x1F4A9, 0x65])
        infos = [(g.codepoint, g.cluster) for g in buf.glyph_infos]
        assert infos == [(0x61, 0), (0x431, 1), (0xE7, 2), (0x1F4A9, 3), (0x65, 4)]

    def test_guess_set_segment_properties(self):
        buf = hb.Buffer()
        buf.add_str("הארץ")

        buf.guess_segment_properties()

        assert buf.direction == "rtl"
        assert buf.script == "Hebr"
        # the guessed language seems to be locale specific
        # assert buf.language == "en-us"
        assert buf.language

        buf.direction = "ltr"
        assert buf.direction == "ltr"

        buf.script = "Latn"
        assert buf.script == "Latn"

        buf.language = "he-il"
        assert buf.language == "he-il"

        buf.set_script_from_ot_tag("mym2")
        assert buf.script == "Mymr"

        buf.set_language_from_ot_tag("BGR")
        assert buf.language == "bg"

    def test_empty_buffer_props(self):
        buf = hb.Buffer()
        assert buf.script == None
        assert buf.language == None
        assert buf.direction == "invalid"

    def test_cluster_level(self):
        buf = hb.Buffer()

        assert buf.cluster_level == hb.BufferClusterLevel.DEFAULT

        buf.cluster_level = hb.BufferClusterLevel.MONOTONE_CHARACTERS
        assert buf.cluster_level == hb.BufferClusterLevel.MONOTONE_CHARACTERS

        buf.cluster_level = hb.BufferClusterLevel.MONOTONE_GRAPHEMES
        assert buf.cluster_level == hb.BufferClusterLevel.MONOTONE_GRAPHEMES

        buf.cluster_level = hb.BufferClusterLevel.CHARACTERS
        assert buf.cluster_level == hb.BufferClusterLevel.CHARACTERS

        buf.cluster_level = hb.BufferClusterLevel.DEFAULT
        assert buf.cluster_level == hb.BufferClusterLevel.DEFAULT

    def test_cluster_level_int(self):
        buf = hb.Buffer()

        assert buf.cluster_level == 0

        buf.cluster_level = 1
        assert buf.cluster_level == 1

        with pytest.raises(ValueError):
            # 5 is not a valid BufferClusterLevel
            buf.cluster_level = 5
        assert buf.cluster_level == 1

    def test_properties(self):
        buf = hb.Buffer()

        assert len(buf) == 0

        assert buf.flags == hb.BufferFlags.DEFAULT
        buf.flags = hb.BufferFlags.BOT | hb.BufferFlags.EOT
        assert buf.flags == hb.BufferFlags.BOT | hb.BufferFlags.EOT

        assert buf.content_type == hb.BufferContentType.INVALID
        buf.content_type = hb.BufferContentType.UNICODE
        assert buf.content_type == hb.BufferContentType.UNICODE

        assert buf.cluster_level == hb.BufferClusterLevel.DEFAULT
        buf.cluster_level = hb.BufferClusterLevel.CHARACTERS
        assert buf.cluster_level == hb.BufferClusterLevel.CHARACTERS

        assert buf.replacement_codepoint == hb.Buffer.DEFAULT_REPLACEMENT_CODEPOINT
        buf.replacement_codepoint = 0
        assert buf.replacement_codepoint == 0

        buf.add_str("ABC")
        assert len(buf) == 3
        buf.clear_contents()
        assert len(buf) == 0
        assert buf.flags == hb.BufferFlags.BOT | hb.BufferFlags.EOT
        buf.reset()
        assert buf.flags == hb.BufferFlags.DEFAULT


class TestBlob:
    def test_from_file_path_fail(self):
        with pytest.raises(hb.HarfBuzzError, match="Failed to open: DOES-NOT-EXIST"):
            blob = hb.Blob.from_file_path("DOES-NOT-EXIST")


class TestFace:
    def test_properties(self, blankfont):
        face = blankfont.face

        assert face.index == 0
        assert face.upem == 1000
        assert face.glyph_count == 9
        assert face.table_tags == [
            "BASE",
            "GPOS",
            "GSUB",
            "OS/2",
            "cmap",
            "cvt ",
            "fpgm",
            "gasp",
            "glyf",
            "head",
            "hhea",
            "hmtx",
            "loca",
            "maxp",
            "name",
            "post",
            "prep",
        ]

        assert face.unicodes == hb.Set(
            {0x61, 0x62, 0x63, 0x64, 0x65, 0xE7, 0x431, 0x1F4A9}
        )
        assert face.variation_selectors == hb.Set()
        assert face.variation_unicodes(1) == hb.Set()


class TestFont:
    def test_get_glyph_extents(self, opensans):
        # <TTGlyph name="A" xMin="0" yMin="0" xMax="1296" yMax="1468">
        extents = opensans.get_glyph_extents(1)
        assert (0, 1468, 1296, -1468) == extents
        assert 0 == extents.x_bearing
        assert 1468 == extents.y_bearing
        assert 1296 == extents.width
        assert -1468 == extents.height
        assert opensans.get_glyph_extents(1000) is None

    def test_get_font_extents(self, blankfont):
        extents = blankfont.get_font_extents("ltr")
        assert (880, -120, 0) == extents
        assert 880 == extents.ascender
        assert -120 == extents.descender
        assert 0 == extents.line_gap
        extents = blankfont.get_font_extents("ttb")
        assert (500, -500, 0) == extents
        assert 500 == extents.ascender
        assert -500 == extents.descender
        assert 0 == extents.line_gap

    def test_get_glyph_name(self, blankfont):
        glyph_name = blankfont.get_glyph_name(1)
        assert glyph_name == "a"
        glyph_name = blankfont.get_glyph_name(1000)
        assert glyph_name is None

    def test_get_nominal_glyph(self, blankfont):
        gid = blankfont.get_nominal_glyph(ord("a"))
        assert gid == 1
        gid = blankfont.get_nominal_glyph(ord("å"))
        assert gid is None

    def test_get_var_coords_normalized(self, mutatorsans):
        coords = mutatorsans.get_var_coords_normalized()
        assert coords == []
        mutatorsans.set_variations({"wght": 500})
        coords = mutatorsans.get_var_coords_normalized()
        assert coords == [0, 0.5]
        mutatorsans.set_variations({"wdth": 1000})
        coords = mutatorsans.get_var_coords_normalized()
        assert coords == [1.0, 0]
        mutatorsans.set_variations({"wdth": 250, "wght": 250})
        coords = mutatorsans.get_var_coords_normalized()
        assert coords == [0.25, 0.25]

    def test_set_var_coords_normalized(self, mutatorsans):
        expected_coords = [0.5, 0.25]
        mutatorsans.set_var_coords_normalized(expected_coords)
        coords = mutatorsans.get_var_coords_normalized()
        assert expected_coords == coords

        expected_coords = [0.5]
        mutatorsans.set_var_coords_normalized(expected_coords)
        coords = mutatorsans.get_var_coords_normalized()
        assert expected_coords == coords

        with pytest.raises(TypeError):
            mutatorsans.set_var_coords_normalized(["a"])

    def test_properties(self, blankfont):
        assert blankfont.scale == (1000, 1000)
        blankfont.scale = (1024, 1024)
        assert blankfont.scale == (1024, 1024)

        assert blankfont.ppem == (0, 0)
        blankfont.ppem = (16, 24)
        assert blankfont.ppem == (16, 24)

        assert blankfont.ptem == 0
        blankfont.ptem = 12.0
        assert blankfont.ptem == 12.0

        assert blankfont.synthetic_slant == 0
        blankfont.synthetic_slant = 0.2
        assert blankfont.synthetic_slant == pytest.approx(0.2)


class TestShape:
    @pytest.mark.parametrize(
        "string, expected",
        [
            ("abcde", [(1, 0), (2, 1), (3, 2), (4, 3), (5, 4)]),
            ("abçde", [(1, 0), (2, 1), (6, 2), (4, 3), (5, 4)]),
            ("aбcde", [(1, 0), (7, 1), (3, 2), (4, 3), (5, 4)]),
            ("abc💩e", [(1, 0), (2, 1), (3, 2), (8, 3), (5, 4)]),
        ],
        ids=["ascii", "latin1", "ucs2", "ucs4"],
    )
    def test_gid_and_cluster_no_features(self, blankfont, string, expected):
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf)
        infos = [(g.codepoint, g.cluster) for g in buf.glyph_infos]
        assert infos == expected

    def test_shape_set_shaper(self, blankfont):
        string = "abcde"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf, shapers=["fallback"])
        pos = [g.position for g in buf.glyph_positions]
        expected = [
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
        ]
        assert pos == expected

    @pytest.mark.skipif(sys.platform != "win32", reason="requires Windows")
    def test_shape_set_shaper_directwrite(self, blankfont):
        string = "abcde"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf, shapers=["directwrite"])
        pos = [g.position for g in buf.glyph_positions]
        expected = [
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
        ]
        assert pos == expected

    @pytest.mark.xfail
    @pytest.mark.skipif(sys.platform != "win32", reason="requires Windows")
    def test_shape_set_shaper_uniscribe(self, blankfont):
        string = "abcde"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf, shapers=["uniscribe"])
        pos = [g.position for g in buf.glyph_positions]
        expected = [
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
        ]
        assert pos == expected

    @pytest.mark.skipif(sys.platform != "darwin", reason="requires macOS")
    def test_shape_set_shaper_coretext(self, blankfont):
        string = "abcde"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf, shapers=["coretext"])
        pos = [g.position for g in buf.glyph_positions]
        expected = [
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
            (0, 0, 0, 0),
        ]
        assert pos == expected

    @pytest.mark.parametrize(
        "string, expected",
        [
            ("abcde", ["a", "b", "c", "d", "e"]),
            ("abçde", ["a", "b", "ccedilla", "d", "e"]),
            ("aбcde", ["a", "uni0431", "c", "d", "e"]),
            ("abc💩e", ["a", "b", "c", "u1F4A9", "e"]),
        ],
        ids=["ascii", "latin1", "ucs2", "ucs4"],
    )
    def test_glyh_name_no_features(self, blankfont, string, expected):
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf)
        # font.get_glyph_name() returns None if the font does not contain glyph names
        # or if the glyph ID does not exist.
        glyph_names = [blankfont.get_glyph_name(g.codepoint) for g in buf.glyph_infos]
        assert glyph_names == expected
        assert blankfont.get_glyph_name(1000) is None
        # font.glyph_to_string() return "gidN" if the font does not contain glyph names
        # or if the glyph ID does not exist.
        glyph_names = [blankfont.glyph_to_string(g.codepoint) for g in buf.glyph_infos]
        assert glyph_names == expected
        assert blankfont.glyph_to_string(1000) == "gid1000"

    @pytest.mark.parametrize(
        "string, features, expected",
        [
            # The calt feature replaces c by a in the context e, d, c', b, a.
            ("edcbaedcba", {}, ["e", "d", "a", "b", "a", "e", "d", "a", "b", "a"]),
            (
                "edcbaedcba",
                {"calt[2]": False},
                ["e", "d", "c", "b", "a", "e", "d", "a", "b", "a"],
            ),
            (
                "edcbaedcba",
                {"calt": [(7, 8, False)]},
                ["e", "d", "a", "b", "a", "e", "d", "c", "b", "a"],
            ),
            (
                "edcbaedcba",
                {"calt": [(0, 10, False), (7, 8, True)]},
                ["e", "d", "c", "b", "a", "e", "d", "a", "b", "a"],
            ),
        ],
    )
    def test_features_slice(self, blankfont, string, features, expected):
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        hb.shape(blankfont, buf, features)

        glyph_names = [blankfont.glyph_to_string(g.codepoint) for g in buf.glyph_infos]
        assert glyph_names == expected


class TestCallbacks:
    def test_nominal_glyph_func(self, blankfont):
        string = "abcde"
        expected = [97, 98, 99, 100, 101]
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()

        def nominal_glyph_func(font, code_point, data):
            return code_point

        funcs = hb.FontFuncs.create()
        funcs.set_nominal_glyph_func(nominal_glyph_func)
        blankfont.funcs = funcs

        hb.shape(blankfont, buf)
        infos = [g.codepoint for g in buf.glyph_infos]
        assert infos == expected

    def test_variation_glyph_func(self, blankfont):
        string = "a\uFE00"
        expected = [ord("a") + 0xFE00]
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()

        def variation_glyph_func(font, unicode, variation_selector, data):
            return unicode + variation_selector

        funcs = hb.FontFuncs.create()
        funcs.set_variation_glyph_func(variation_glyph_func)
        blankfont.funcs = funcs

        hb.shape(blankfont, buf)
        infos = [g.codepoint for g in buf.glyph_infos]
        assert infos == expected

    def test_glyph_h_advance_func(self, blankfont):
        string = "abcde"
        expected = [456, 456, 456, 456, 456]
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()

        def h_advance_func(font, gid, data):
            return 456

        funcs = hb.FontFuncs.create()
        funcs.set_glyph_h_advance_func(h_advance_func)
        blankfont.funcs = funcs

        hb.shape(blankfont, buf)
        infos = [pos.x_advance for pos in buf.glyph_positions]
        assert infos == expected

    def test_glyph_v_metrics_funcs(self, blankfont):
        string = "abcde"
        expected = [(456, -345, -567)] * 5
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        buf.direction = "TTB"

        def v_advance_func(font, gid, data):
            return 456

        def v_origin_func(font, gid, data):
            return (True, 345, 567)

        funcs = hb.FontFuncs.create()
        funcs.set_glyph_v_advance_func(v_advance_func)
        funcs.set_glyph_v_origin_func(v_origin_func)
        blankfont.funcs = funcs

        hb.shape(blankfont, buf)
        infos = [
            (pos.y_advance, pos.x_offset, pos.y_offset) for pos in buf.glyph_positions
        ]
        assert infos == expected

    def test_font_extents_funcs(self, blankfont):
        def font_h_extents_func(font, data):
            return hb.FontExtents(123, -456, 789)

        def font_v_extents_func(font, data):
            return hb.FontExtents(987, -654, 321)

        funcs = hb.FontFuncs.create()
        funcs.set_font_h_extents_func(font_h_extents_func)
        funcs.set_font_v_extents_func(font_v_extents_func)
        blankfont.funcs = funcs
        assert (123, -456, 789) == blankfont.get_font_extents("ltr")
        assert (987, -654, 321) == blankfont.get_font_extents("ttb")

    def test_message_func(self, blankfont):
        # Glyph IDs 1, 2, 3, 4, 5 map to glyphs a, b, c, d, e.
        # The calt feature replaces c by a in the context e, d, c', b, a.
        # The kern feature kerns b, a by +100.
        string = "edcba"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()

        messages = []
        infos_trace = []
        positions_trace = []

        def message(msg):
            messages.append(msg)
            infos_trace.append(buf.glyph_infos)
            positions_trace.append(buf.glyph_positions)

        buf.set_message_func(message)
        hb.shape(blankfont, buf)
        gids = [g.codepoint for g in buf.glyph_infos]
        assert gids == [5, 4, 1, 2, 1]
        pos = [g.x_advance for g in buf.glyph_positions]
        assert pos == [0, 0, 0, 100, 0]
        expected_messages = [
            "start table GSUB script tag 'DFLT'",
            "start lookup 0 feature 'calt'",
            "recursing to lookup 1 at 2",
            "replacing glyph at 2 (single substitution)",
            "replaced glyph at 2 (single substitution)",
            "recursed to lookup 1",
            "end lookup 0 feature 'calt'",
            "end table GSUB script tag 'DFLT'",
            "start table GPOS script tag 'DFLT'",
            "start lookup 0 feature 'kern'",
            "try kerning glyphs at 3,4",
            "kerned glyphs at 3,4",
            "tried kerning glyphs at 3,4",
            "end lookup 0 feature 'kern'",
            "end table GPOS script tag 'DFLT'",
        ]
        assert messages == expected_messages
        gids_trace = [[g.codepoint for g in infos] for infos in infos_trace]
        assert gids_trace == [
            [5, 4, 3, 2, 1],
            [5, 4, 3, 2, 1],
            [5, 4, 3, 2, 1],
            [5, 4, 3, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
            [5, 4, 1, 2, 1],
        ]
        advances_trace = [[g.x_advance for g in pos] for pos in positions_trace if pos]
        assert advances_trace == [
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 100, 0],
            [0, 0, 0, 100, 0],
            [0, 0, 0, 100, 0],
            [0, 0, 0, 100, 0],
        ]

    def test_message_func_return_false(self, blankfont):
        # Glyph IDs 1, 2, 3, 4, 5 map to glyphs a, b, c, d, e.
        # The calt feature replaces c by a in the context e, d, c', b, a.
        # The kern feature kerns b, a by +100.
        string = "edcba"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()

        messages = []
        infos_trace = []
        positions_trace = []

        def message(msg):
            messages.append(msg)
            infos_trace.append(buf.glyph_infos)
            positions_trace.append(buf.glyph_positions)
            return False

        buf.set_message_func(message)
        hb.shape(blankfont, buf)
        gids = [g.codepoint for g in buf.glyph_infos]
        assert gids == [5, 4, 3, 2, 1]
        pos = [g.x_advance for g in buf.glyph_positions]
        assert pos == [0, 0, 0, 0, 0]
        expected_messages = [
            "start table GSUB script tag 'DFLT'",
            "start table GPOS script tag 'DFLT'",
        ]
        assert messages == expected_messages
        gids_trace = [[g.codepoint for g in infos] for infos in infos_trace]
        assert gids_trace == [[5, 4, 3, 2, 1], [5, 4, 3, 2, 1]]
        advances_trace = [[g.x_advance for g in pos] for pos in positions_trace if pos]
        assert advances_trace == [[0, 0, 0, 0, 0]]

    def test_message_func_crash(self, blankfont):
        string = "edcba"
        buf = hb.Buffer()
        buf.add_str(string)
        buf.guess_segment_properties()
        message_collector = MessageCollector()
        buf.set_message_func(message_collector.message)
        hb.shape(blankfont, buf)

    def test_draw_funcs(self, opensans):
        funcs = hb.DrawFuncs()
        container = []

        def move_to(x, y, c):
            c.append(f"M{x:g},{y:g}")

        def line_to(x, y, c):
            c.append(f"L{x:g},{y:g}")

        def cubic_to(c1x, c1y, c2x, c2y, x, y, c):
            c.append(f"C{c1x:g},{c1y:g} {c2x:g},{c2y:g} {x:g},{y:g}")

        def quadratic_to(c1x, c1y, x, y, c):
            c.append(f"Q{c1x:g},{c1y:g} {x:g},{y:g}")

        def close_path(c):
            c.append("Z")

        funcs.set_move_to_func(move_to)
        funcs.set_line_to_func(line_to)
        funcs.set_cubic_to_func(cubic_to)
        funcs.set_quadratic_to_func(quadratic_to)
        funcs.set_close_path_func(close_path)
        opensans.draw_glyph(1, funcs, container)
        assert (
            "".join(container)
            == "M1120,0L938,465L352,465L172,0L0,0L578,1468L721,1468L1296,0L1120,0ZM885,618L715,1071Q682,1157 647,1282Q625,1186 584,1071L412,618L885,618Z"
        )

    @pytest.mark.xfail(
        platform.python_implementation() == "PyPy",
        reason="PyPy's ctypes has no 'pythonapi' attribute",
    )
    def test_draw_funcs_pycapsule(self, opensans):
        import ctypes
        import uharfbuzz._harfbuzz_test

        funcs = hb.DrawFuncs()

        PyCapsule_New = ctypes.pythonapi.PyCapsule_New
        PyCapsule_New.restype = ctypes.py_object
        PyCapsule_New.argtypes = (ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p)

        def cap(x):
            return PyCapsule_New(x, None, None)

        lib = ctypes.cdll.LoadLibrary(uharfbuzz._harfbuzz_test.__file__)
        container = ctypes.create_string_buffer(1000)
        container_cap = cap(container)

        funcs.set_move_to_func(cap(lib._test_move_to))
        funcs.set_line_to_func(cap(lib._test_line_to))
        funcs.set_cubic_to_func(cap(lib._test_cubic_to))
        funcs.set_quadratic_to_func(cap(lib._test_quadratic_to))
        funcs.set_close_path_func(cap(lib._test_close_path))
        opensans.draw_glyph(1, funcs, container_cap)

        assert (
            container.value
            == b"M1120,0L938,465L352,465L172,0L0,0L578,1468L721,1468L1296,0L1120,0ZM885,618L715,1071Q682,1157 647,1282Q625,1186 584,1071L412,618L885,618Z"
        )

    def test_draw_pen(self, opensans):
        class TestPen:
            def __init__(self):
                self.value = []

            def moveTo(self, p0):
                self.value.append(("moveTo", (p0,)))

            def lineTo(self, p1):
                self.value.append(("lineTo", (p1,)))

            def qCurveTo(self, *points):
                self.value.append(("qCurveTo", points))

            def curveTo(self, *points):
                self.value.append(("curveTo", points))

            def closePath(self):
                self.value.append(("closePath", ()))

        pen = TestPen()
        opensans.draw_glyph_with_pen(1, pen)
        assert pen.value == [
            ("moveTo", ((1120, 0),)),
            ("lineTo", ((938, 465),)),
            ("lineTo", ((352, 465),)),
            ("lineTo", ((172, 0),)),
            ("lineTo", ((0, 0),)),
            ("lineTo", ((578, 1468),)),
            ("lineTo", ((721, 1468),)),
            ("lineTo", ((1296, 0),)),
            ("lineTo", ((1120, 0),)),
            ("closePath", ()),
            ("moveTo", ((885, 618),)),
            ("lineTo", ((715, 1071),)),
            ("qCurveTo", ((682, 1157), (647, 1282))),
            ("qCurveTo", ((625, 1186), (584, 1071))),
            ("lineTo", ((412, 618),)),
            ("lineTo", ((885, 618),)),
            ("closePath", ()),
        ]


class TestPaintFuncs:
    @staticmethod
    def setup_funcs():
        def push_transform_func(xx, xy, yx, yy, dx, dy, conainer):
            conainer.append(
                f"start transform "
                f"{xx:.3g} {xy:.3g} {yx:.3g} {yy:.3g} {dx:.3g} {dy:.3g}"
            )
            conainer.level += 1

        def pop_transform_func(conainer):
            conainer.level -= 1
            conainer.append(f"end transform")

        def color_glyph_func(gid, conainer):
            conainer.append(f"paint color glyph {gid}; acting as failed")
            return False

        def push_clip_glyph_func(gid, conainer):
            conainer.append(f"start clip glyph {gid}")
            conainer.level += 1
            return True

        def push_clip_rectangle_func(xmin, ymin, xmax, ymax, conainer):
            conainer.append(
                f"start clip rectangle {xmin:.3g} {ymin:.3g} {xmax:.3g} {ymax:.3g}"
            )
            conainer.level += 1

        def pop_clip_func(conainer):
            conainer.level -= 1
            conainer.append(f"end clip")

        def color_func(color, is_foreground, conainer):
            r, g, b, a = color
            conainer.append(f"solid {r} {g} {b} {a}")

        def image_func(image, width, height, format, slant, extents, conainer):
            conainer.append(
                f"image type {format} "
                f"size {width} {height} "
                f"slant {slant:.3g} "
                f"extents {extents.x_bearing} {extents.y_bearing} {extents.width} {extents.height}"
            )
            return True

        def linear_gradient_func(color_line, x0, y0, x1, y1, x2, y2, conainer):
            conainer.append(f"linear gradient")
            conainer.level += 1
            conainer.append(f"p0 {x0:.3g} {y0:.3g}")
            conainer.append(f"p1 {x1:.3g} {y1:.3g}")
            conainer.append(f"p2 {x2:.3g} {y2:.3g}")
            conainer.append_color_line(color_line)
            conainer.level -= 1

        def radial_gradient_func(color_line, x0, y0, r0, x1, y1, r1, conainer):
            conainer.append(f"radial gradient")
            conainer.level += 1
            conainer.append(f"p0 {x0:.3g} {y0:.3g} radius {r0:.3g}")
            conainer.append(f"p1 {x1:.3g} {y1:.3g} radius {r1:.3g}")
            conainer.append_color_line(color_line)
            conainer.level -= 1

        def sweep_gradient_func(color_line, cx, cy, start_angle, end_angle, conainer):
            conainer.append(f"sweep gradient")
            conainer.level += 1
            conainer.append(f"center {cx:.3g} {cy:.3g}")
            conainer.append(f"angles {start_angle:.3g} {end_angle:.3g}")
            conainer.append_color_line(color_line)
            conainer.level -= 1

        def push_group_func(conainer):
            conainer.append(f"push group")
            conainer.level += 1

        def pop_group_func(mode, conainer):
            conainer.level -= 1
            conainer.append(f"pop group mode {int(mode)}")

        def custom_palette_color_func(color_index, conainer):
            return None

        funcs = hb.PaintFuncs()
        funcs.set_push_transform_func(push_transform_func)
        funcs.set_pop_transform_func(pop_transform_func)
        funcs.set_color_glyph_func(color_glyph_func)
        funcs.set_push_clip_glyph_func(push_clip_glyph_func)
        funcs.set_push_clip_rectangle_func(push_clip_rectangle_func)
        funcs.set_pop_clip_func(pop_clip_func)
        funcs.set_color_func(color_func)
        funcs.set_image_func(image_func)
        funcs.set_linear_gradient_func(linear_gradient_func)
        funcs.set_radial_gradient_func(radial_gradient_func)
        funcs.set_sweep_gradient_func(sweep_gradient_func)
        funcs.set_push_group_func(push_group_func)
        funcs.set_pop_group_func(pop_group_func)
        funcs.set_custom_palette_color_func(custom_palette_color_func)

        class Container:
            def __init__(self):
                self.level = 0
                self.lines = []

            def append(self, line):
                indent = " " * self.level * 2
                self.lines.append(indent + line)

            def append_color_line(self, color_line):
                self.append(f"colors {int(color_line.extend)}")

                self.level += 1
                for stop in color_line.color_stops:
                    r, g, b, a = stop.color
                    self.append(f"{stop.offset:.3g} {r} {g} {b} {a}")
                self.level -= 1

            def value(self):
                return "\n".join(self.lines)

        return funcs, Container()

    @pytest.mark.parametrize(
        "fontpath, glyph, expectedpath",
        [
            ("noto_handwriting-cff2_colr_1.otf", 10, "hand-10"),
            ("test_glyphs-glyf_colr_1.ttf", 6, "test-6"),
            ("test_glyphs-glyf_colr_1.ttf", 10, "test-10"),
            ("test_glyphs-glyf_colr_1.ttf", 92, "test-92"),
            ("test_glyphs-glyf_colr_1.ttf", 106, "test-106"),
            ("test_glyphs-glyf_colr_1.ttf", 116, "test-116"),
            ("test_glyphs-glyf_colr_1.ttf", 123, "test-123"),
            ("test_glyphs-glyf_colr_1.ttf", 154, "test-154"),
            ("test_glyphs-glyf_colr_1.ttf", 165, "test-165"),
            ("test_glyphs-glyf_colr_1.ttf", 175, "test-175"),
        ],
    )
    def test_paint(self, fontpath, glyph, expectedpath):
        blob = hb.Blob.from_file_path(TESTDATA / fontpath)
        face = hb.Face(blob)
        font = hb.Font(face)

        funcs, container = self.setup_funcs()
        font.paint_glyph(glyph, funcs, container)
        result = container.value()

        with open(TESTDATA / "expected" / expectedpath) as f:
            expected = "".join(line for line in f.readlines() if line[0] != "#")

        assert result.strip() == expected.strip()


class MessageCollector:
    def message(self, message):
        pass


class TestOTLayout:
    # The test font contains a BASE table with some test values
    def test_ot_layout_get_baseline_invalid_tag(self, blankfont):
        with pytest.raises(ValueError):
            # invalid baseline tag
            baseline = hb.ot_layout_get_baseline(blankfont, "xxxx", "LTR", "", "")

    @pytest.mark.parametrize(
        "baseline_tag, script_tag, direction, expected_value",
        [
            ("icfb", "grek", "LTR", None),  # BASE table doesn't contain grek script
            ("icfb", "latn", "LTR", -70),
            ("icft", "latn", "LTR", 830),
            ("romn", "latn", "LTR", 0),
            ("ideo", "latn", "LTR", -120),
            ("icfb", "kana", "LTR", -71),
            ("icft", "kana", "LTR", 831),
            ("romn", "kana", "LTR", 1),
            ("ideo", "kana", "LTR", -121),
            ("icfb", "latn", "TTB", 50),
            ("icft", "latn", "TTB", 950),
            ("romn", "latn", "TTB", 120),
            ("ideo", "latn", "TTB", 0),
            ("icfb", "kana", "TTB", 51),
            ("icft", "kana", "TTB", 951),
            ("romn", "kana", "TTB", 121),
            ("ideo", "kana", "TTB", 1),
        ],
    )
    def test_ot_layout_get_baseline(
        self, blankfont, baseline_tag, script_tag, direction, expected_value
    ):
        value = hb.ot_layout_get_baseline(
            blankfont, baseline_tag, direction, script_tag, ""
        )
        assert value == expected_value

    def test_ot_layout_language_get_feature_tags(self, blankfont):
        tags = hb.ot_layout_language_get_feature_tags(blankfont.face, "GPOS")
        assert tags == ["kern"]
        tags = hb.ot_layout_language_get_feature_tags(blankfont.face, "GSUB")
        assert tags == ["calt"]

    def test_ot_layout_table_get_script_tags(self, blankfont):
        tags = hb.ot_layout_table_get_script_tags(blankfont.face, "GPOS")
        assert tags == ["DFLT"]

    def test_ot_layout_script_get_language_tags(self, blankfont):
        tags = hb.ot_layout_script_get_language_tags(blankfont.face, "GPOS", 0)
        assert tags == []

    def test_ot_layout_lookup_get_glyph_alternates(self, blankfont):
        gid = blankfont.get_nominal_glyph(ord("c"))
        alternates = hb.ot_layout_lookup_get_glyph_alternates(blankfont.face, 1, gid)
        assert alternates == [1]

    def test_ot_layout_has_glyph_classes(self, opensans):
        assert hb.ot_layout_has_glyph_classes(opensans.face)

    def test_ot_layout_has_no_glyph_classes(self, blankfont):
        assert hb.ot_layout_has_glyph_classes(blankfont.face) == False

    def test_ot_layout_get_glyph_class(self, opensans):
        glyph_class = hb.ot_layout_get_glyph_class(opensans.face, 1)
        assert glyph_class == hb.OTLayoutGlyphClass.BASE_GLYPH

    def test_ot_layout_has_positioning(self, opensans):
        assert hb.ot_layout_has_positioning(opensans.face)

    def test_ot_layout_has_no_positioning(self, mathfont):
        assert hb.ot_layout_has_positioning(mathfont.face) == False

    def test_ot_layout_has_substitution(self, opensans):
        assert hb.ot_layout_has_substitution(opensans.face)

    def test_ot_layout_has_no_substitution(self, mathfont):
        assert hb.ot_layout_has_substitution(mathfont.face) == False

    def test_ot_tag_to_script(self):
        assert hb.ot_tag_to_script("mym2") == "Mymr"

    def test_ot_tag_to_language(self):
        assert hb.ot_tag_to_language("BGR") == "bg"


class TestOTColor:
    def test_ot_color_has_palettes(self, colorv0font):
        assert hb.ot_color_has_palettes(colorv0font.face)

    def test_ot_color_has_no_palettes(self, blankfont):
        assert hb.ot_color_has_palettes(blankfont.face) == False

    def test_ot_color_palette_get_count(self, colorv0font, colorv1font):
        assert hb.ot_color_palette_get_count(colorv0font.face) == 1
        assert hb.ot_color_palette_get_count(colorv1font.face) == 3

    def test_ot_color_palette_get_colors(self, colorv0font, colorv1font):
        palette = hb.ot_color_palette_get_colors(colorv0font.face, 0)
        assert palette == [(200, 0, 0, 255)]

        palette = hb.ot_color_palette_get_colors(colorv1font.face, 0)
        assert palette == [
            (255, 0, 0, 255),
            (255, 165, 0, 255),
            (255, 255, 0, 255),
            (0, 128, 0, 255),
            (0, 0, 255, 255),
            (75, 0, 130, 255),
            (238, 130, 238, 255),
            (250, 240, 230, 255),
            (47, 79, 79, 255),
            (255, 255, 255, 255),
            (0, 0, 0, 255),
            (104, 199, 232, 255),
            (255, 220, 1, 255),
            (128, 128, 128, 255),
        ]

    def test_ot_color_palette_get_colors_no_palettes(self, blankfont):
        palette = hb.ot_color_palette_get_colors(blankfont.face, 0)
        assert palette == []

    def test_ot_color_palette_get_flags(self, colorv1font):
        flags = hb.ot_color_palette_get_flags(colorv1font.face, 0)
        assert flags == hb.OTColorPaletteFlags.DEFAULT

        flags = hb.ot_color_palette_get_flags(colorv1font.face, 1)
        assert flags == hb.OTColorPaletteFlags.USABLE_WITH_DARK_BACKGROUND

        flags = hb.ot_color_palette_get_flags(colorv1font.face, 2)
        assert flags == hb.OTColorPaletteFlags.USABLE_WITH_LIGHT_BACKGROUND

    def test_ot_color_palette_get_flags_no_palettes(self, colorv0font):
        flags = hb.ot_color_palette_get_flags(colorv0font.face, 0)
        assert flags == hb.OTColorPaletteFlags.DEFAULT

    def test_ot_color_palette_get_name_id(self, colorv1font):
        name_id = hb.ot_color_palette_get_name_id(colorv1font.face, 0)
        assert name_id == None
        name_id = hb.ot_color_palette_get_name_id(colorv1font.face, 1)
        assert name_id == None
        name_id = hb.ot_color_palette_get_name_id(colorv1font.face, 2)
        assert name_id == None

    def test_ot_color_palette_get_name_id_no_palettes(self, blankfont):
        name_id = hb.ot_color_palette_get_name_id(blankfont.face, 0)
        assert name_id == None

    def test_ot_color_palette_color_get_name_id(self, colorv1font):
        name_id = hb.ot_color_palette_color_get_name_id(colorv1font.face, 0)
        assert name_id == None
        name_id = hb.ot_color_palette_color_get_name_id(colorv1font.face, 1)
        assert name_id == None
        name_id = hb.ot_color_palette_color_get_name_id(colorv1font.face, 2)
        assert name_id == None

    def test_ot_color_has_layers(self, colorv0font):
        assert hb.ot_color_has_layers(colorv0font.face)

    def test_ot_color_has_no_layers(self, blankfont):
        assert hb.ot_color_has_layers(blankfont.face) == False

    def test_ot_color_glyph_get_layers(self, colorv0font):
        layers = hb.ot_color_glyph_get_layers(colorv0font.face, 1)
        assert layers == [(1, 0)]

        layers = hb.ot_color_glyph_get_layers(colorv0font.face, 0)
        assert layers == []

    def test_ot_color_glyph_get_layers_no_layers(self, blankfont):
        layers = hb.ot_color_glyph_get_layers(blankfont.face, 1)
        assert layers == []

    def test_ot_color_has_paint(self, colorv1font):
        assert hb.ot_color_has_paint(colorv1font.face)

    def test_ot_color_has_no_paint(self, blankfont):
        assert hb.ot_color_has_paint(blankfont.face) == False

    def test_ot_color_glyph_has_paint(self, colorv1font):
        assert hb.ot_color_glyph_has_paint(colorv1font.face, 1) == False
        assert hb.ot_color_glyph_has_paint(colorv1font.face, 7) == True

    def test_ot_color_glyph_has_paint_no_paint(self, blankfont):
        assert hb.ot_color_glyph_has_paint(blankfont.face, 1) == False

    def test_ot_color_has_svg(self, blankfont):
        assert hb.ot_color_has_svg(blankfont.face) == False

        blob = hb.ot_color_glyph_get_svg(blankfont.face, 1)
        assert len(blob) == 0

    def test_ot_color_has_png(self, blankfont):
        assert hb.ot_color_has_png(blankfont.face) == False

        blob = hb.ot_color_glyph_get_png(blankfont, 1)
        assert len(blob) == 0


class TestOTMath:
    def test_ot_math_has_data(self, mathfont):
        assert hb.ot_math_has_data(mathfont.face)

    def test_ot_math_has_no_data(self, blankfont):
        assert hb.ot_math_has_data(blankfont.face) == False

    @pytest.mark.parametrize(
        "constant, expected",
        [
            (hb.OTMathConstant.SCRIPT_PERCENT_SCALE_DOWN, 70),
            (hb.OTMathConstant.SCRIPT_SCRIPT_PERCENT_SCALE_DOWN, 55),
            (hb.OTMathConstant.RADICAL_KERN_BEFORE_DEGREE, 65),
            (hb.OTMathConstant.RADICAL_KERN_AFTER_DEGREE, -335),
            (hb.OTMathConstant.RADICAL_DEGREE_BOTTOM_RAISE_PERCENT, 55),
        ],
    )
    def test_ot_math_get_constant(self, mathfont, constant, expected):
        assert hb.ot_math_get_constant(mathfont, constant) == expected

    def test_ot_math_get_constant_invalid(self, mathfont):
        with pytest.raises(ValueError):
            hb.ot_math_get_constant(mathfont, 1000)

    @pytest.mark.parametrize(
        "glyph, expected",
        [
            ("uni222B", 230),
            ("uni210B", 40),
            ("uni222B.dsp", 540),
            ("u1D435", 30),
            ("A", 0),
        ],
    )
    def test_ot_math_get_glyph_italics_correction(self, mathfont, glyph, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assert hb.ot_math_get_glyph_italics_correction(mathfont, gid) == expected

    @pytest.mark.parametrize(
        "glyph, expected",
        [
            ("u1D435", 380),
            ("A", 360),
            ("parenleft", 250),
        ],
    )
    def test_ot_math_get_glyph_top_accent_attachment(self, mathfont, glyph, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assert hb.ot_math_get_glyph_top_accent_attachment(mathfont, gid) == expected

    @pytest.mark.parametrize(
        "glyph, expected",
        [
            ("A", False),
            ("parenleft", True),
            ("parenright", True),
            ("bar", True),
            ("uni221A", True),
        ],
    )
    def test_ot_math_is_glyph_extended_shape(self, mathfont, glyph, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assert hb.ot_math_is_glyph_extended_shape(mathfont.face, gid) == expected

    @pytest.mark.parametrize(
        "direction, expected",
        [
            ("LTR", 100),
            ("RTL", 100),
            ("TTB", 100),
            ("BTT", 100),
        ],
    )
    def test_ot_math_get_min_connector_overlap(self, mathfont, direction, expected):
        assert hb.ot_math_get_min_connector_overlap(mathfont, direction) == expected

    @pytest.mark.parametrize(
        "glyph, kern, height, expected",
        [
            ("A", hb.OTMathKern.TOP_RIGHT, 250, 0),
            ("A", hb.OTMathKern.TOP_RIGHT, 300, -18),
            ("A", hb.OTMathKern.TOP_RIGHT, 400, -66),
            ("F", hb.OTMathKern.BOTTOM_RIGHT, 0, -200),
            ("F", hb.OTMathKern.BOTTOM_RIGHT, 150, -44),
            ("F", hb.OTMathKern.BOTTOM_RIGHT, 300, 44),
            ("J", hb.OTMathKern.TOP_LEFT, 250, 64),
            ("J", hb.OTMathKern.TOP_LEFT, 300, -28),
            ("J", hb.OTMathKern.TOP_LEFT, 400, -28),
            ("M", hb.OTMathKern.BOTTOM_LEFT, 0, 40),
            ("M", hb.OTMathKern.BOTTOM_LEFT, 150, 40),
            ("M", hb.OTMathKern.BOTTOM_LEFT, 300, 40),
        ],
    )
    def test_ot_math_get_glyph_kerning(self, mathfont, glyph, kern, height, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assert hb.ot_math_get_glyph_kerning(mathfont, gid, kern, height) == expected

    def test_ot_math_get_glyph_kerning_invalid(self, mathfont):
        gid = mathfont.get_glyph_from_name("A")
        with pytest.raises(ValueError):
            hb.ot_math_get_glyph_kerning(mathfont, gid, 1000, 100)

    @pytest.mark.parametrize(
        "glyph, kern, expected",
        [
            (
                "u1D434",
                hb.OTMathKern.TOP_RIGHT,
                [(213, 58), (350, -58), (0x7FFFFFFF, -70)],
            ),
            (
                "u1D435",
                hb.OTMathKern.BOTTOM_RIGHT,
                [(160, -50), (283, -12), (0x7FFFFFFF, 20)],
            ),
            (
                "u1D435",
                hb.OTMathKern.TOP_LEFT,
                [(176, 81), (0x7FFFFFFF, -50)],
            ),
            (
                "U",
                hb.OTMathKern.BOTTOM_LEFT,
                [(126, -80), (256, -60), (0x7FFFFFFF, 36)],
            ),
        ],
    )
    def test_ot_math_get_glyph_kernings(self, mathfont, glyph, kern, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assert hb.ot_math_get_glyph_kernings(mathfont, gid, kern) == expected

    def test_ot_math_get_glyph_kernings_invalid(self, mathfont):
        gid = mathfont.get_glyph_from_name("A")
        with pytest.raises(ValueError):
            hb.ot_math_get_glyph_kernings(mathfont, gid, 1000)

    @pytest.mark.parametrize(
        "glyph, direction, expected",
        [
            (
                "parenleft",
                "TTB",
                [
                    ("parenleft", 933),
                    ("parenleft.s1", 1187),
                    ("parenleft.s2", 1427),
                    ("parenleft.s3", 1667),
                    ("parenleft.s4", 1907),
                    ("parenleft.s5", 2145),
                    ("parenleft.s6", 2385),
                    ("parenleft.s7", 2625),
                    ("parenleft.s8", 2865),
                    ("parenleft.s9", 3101),
                    ("parenleft.s10", 3341),
                    ("parenleft.s11", 3581),
                    ("parenleft.s12", 3821),
                ],
            ),
            ("parenleft", "LTR", []),
            ("uni2211", "TTB", [("uni2211", 1031), ("uni2211.s1", 1326)]),
            (
                "uni0302",
                "LTR",
                [
                    ("uni0302", 283),
                    ("circumflex.s1", 574),
                    ("circumflex.s2", 1003),
                    ("circumflex.s3", 1496),
                    ("circumflex.s4", 1932),
                    ("circumflex.s5", 2385),
                ],
            ),
            ("uni0302", "TTB", []),
        ],
    )
    def test_ot_math_get_glyph_variants(self, mathfont, glyph, direction, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        variants = hb.ot_math_get_glyph_variants(mathfont, gid, direction)
        result = [(mathfont.get_glyph_name(v.glyph), v.advance) for v in variants]
        assert result == expected

    @pytest.mark.parametrize(
        "glyph, direction, expected",
        [
            (
                "parenleft",
                "TTB",
                (
                    [
                        ("uni239D", 0, 250, 1273, 0),
                        ("uni239C", 1000, 1000, 1252, hb.OTMathGlyphPartFlags.EXTENDER),
                        ("uni239B", 250, 0, 1273, 0),
                    ],
                    0,
                ),
            ),
            ("parenleft", "LTR", ([], 0)),
            (
                "uni222B",
                "TTB",
                (
                    [
                        ("uni2321", 0, 800, 1896, 0),
                        (
                            "integral.x",
                            600,
                            600,
                            1251,
                            hb.OTMathGlyphPartFlags.EXTENDER,
                        ),
                        ("uni2320", 800, 0, 1630, 0),
                    ],
                    80,
                ),
            ),
            (
                "uni23DE",
                "LTR",
                (
                    [
                        ("uni23DE.l", 0, 300, 957, 0),
                        ("uni23B4.x", 600, 600, 751, hb.OTMathGlyphPartFlags.EXTENDER),
                        ("uni23DE.m", 300, 300, 943, 0),
                        ("uni23B4.x", 600, 600, 751, hb.OTMathGlyphPartFlags.EXTENDER),
                        ("uni23DE.r", 300, 0, 957, 0),
                    ],
                    0,
                ),
            ),
            ("uni0302", "TTB", ([], 0)),
        ],
    )
    def test_ot_math_get_glyph_assembly(self, mathfont, glyph, direction, expected):
        gid = mathfont.get_glyph_from_name(glyph)
        assembly, italics_correction = hb.ot_math_get_glyph_assembly(
            mathfont, gid, direction
        )
        result = [
            (
                mathfont.get_glyph_name(v.glyph),
                v.start_connector_length,
                v.end_connector_length,
                v.full_advance,
                v.flags,
            )
            for v in assembly
        ]
        assert (result, italics_correction) == expected


def test_harfbuzz_version():
    v = hb.version_string()
    assert isinstance(v, str)


def test_uharfbuzz_version():
    v = hb.__version__
    assert isinstance(v, str)
    assert "unknown" not in v


def test_create_sub_font():
    blob = hb.Blob.from_file_path(ADOBE_BLANK_TTF_PATH)
    face = hb.Face(blob)
    font = hb.Font(face)
    font2 = hb.Font(font)
    assert font is not font2
    assert font.face is font2.face


def test_harfbuzz_repacker():
    table_data = [
        bytes(b"\x00\x00\xff\xff\x00\x01\x00\x00"),
        bytes(b"\x00\x00\x00\x00"),
        bytes(b"\x00\x01latn\x00\x00"),
        bytes(b"\x00\x00\x00\x01\x00\x01"),
        bytes(b"\x00\x01test\x00\x00"),
        bytes(b"\x00\x01\x00\x01\x00\x02"),
        bytes(b"\x00\x01\x00\x00\x00\x01"),
        bytes(b"\x00\x01\x00\x00\x00\x01\x00\x00"),
        bytes(b"\x00\x01\x00\x01\x00\x01"),
        bytes(b"\x00\x02\x00\x01\x00\x02\x00\x01\x00\x00"),
        bytes(b"\x00\x01\x00\x00"),
        bytes(b"\x00\x01\x00\x00\x00\x01\x00\x00"),
        bytes(b"\x00\x05\x00\x00\x00\x01\x00\x00"),
        bytes(b"\x00\x02\x00\x00\x00\x00"),
        bytes(b"\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"),
    ]
    obj_list = [
        ([], []),
        ([(0, 2, 1)], []),
        ([(6, 2, 2)], []),
        ([], []),
        ([(6, 2, 4)], []),
        ([], []),
        ([(2, 2, 6)], []),
        ([(6, 2, 7)], []),
        ([], []),
        ([], []),
        ([(2, 2, 10)], []),
        ([(2, 2, 9), (6, 2, 11)], []),
        ([(6, 2, 12)], []),
        ([(2, 2, 8), (4, 2, 13)], []),
        ([(4, 2, 3), (6, 2, 5), (8, 2, 14)], []),
    ]
    expected_data = bytes(
        b'\x00\x01\x00\x00\x00\x10\x00\x18\x00\n\x00\x02\x00\x1a\x00"\x00\x01latn\x00\x10\x00\x01test\x00\x1c\x00\x1a\x00\x00\x00\x01\x00\x00\x00\x01\x00\x1e\x00\x05\x00\x00\x00\x01\x00\x1c\x00\x00\x00\x01\x00\x01\x00\x00\xff\xff\x00\x01\x00\x00\x00\x01\x00\x0e\x00\x01\x00\x01\x00\x12\x00\x01\x00\x0e\x00\x01\x00\x01\x00\x02\x00\x01\x00\n\x00\x01\x00\x01\x00\x01\x00\x02\x00\x01\x00\x02\x00\x01\x00\x00'
    )
    packed_data = hb.repack(table_data, obj_list)
    assert expected_data == packed_data


@pytest.mark.skipif(sys.platform != "darwin", reason="requires macOS")
def test_sparsefont_coretext(sparsefont):
    buf = hb.Buffer()
    buf.add_str("ABC")
    buf.guess_segment_properties()
    with pytest.raises(RuntimeError):
        hb.shape(sparsefont, buf, shapers=["coretext"])


def test_set():
    s1 = hb.Set()
    s2 = hb.Set({1, 3, 4})
    s3 = hb.Set([1, 3, 4])

    assert s1 != None
    assert s1 != False
    assert not s1
    assert s2
    assert s1 != s2
    assert not s1 == s2
    assert not s2 != s3
    assert s2 == s3

    s1.add(1)
    s1.add_range(3, 4)
    assert s1 == s2

    assert -1 not in s1
    assert 1 in s1
    s1.remove(1)
    assert 1 not in s1

    assert list(s1) == [3, 4]

    s1 &= hb.Set({3, 4, 5})
    assert list(s1) == [3, 4]
    s1 -= hb.Set({3})
    assert list(s1) == [4]
    s1 ^= hb.Set({5})
    assert list(s1) == [4, 5]
    s1 |= {8}  # Update accepts set() as well

    assert list(s1) == [4, 5, 8]
    assert len(s1) == 3

    assert s1.min == 4
    assert s1.max == 8

    assert repr(s1) == "Set({4, 5, 8})"
    s1.invert()
    assert repr(s1) == "Set({...})"

    iter(iter(hb.Set({})))


def test_map():
    m1 = hb.Map()
    m2 = hb.Map({1: 2, 3: 4})
    m3 = hb.Map({1: 2, 3: 4})

    assert m1 != None
    assert m1 != False
    assert not m1
    assert m2
    assert m1 != m2
    assert not m1 == m2
    assert not m2 != m3
    assert m2 == m3
    m1[1] = 2
    assert m1[1] == 2
    assert m1 != m2
    m1[3] = 4
    assert m1 == m2

    assert -1 not in m2
    assert 1 in m1
    del m1[1]
    assert 1 not in m1
    assert len(m1) == 1

    assert set(m2.items()) == {(1, 2), (3, 4)}
    assert set(m2.keys()) == {1, 3}
    assert set(m2) == {1, 3}
    assert set(m2.values()) == {2, 4}

    m4 = hb.Map(m3)
    m5 = hb.Map()
    m5.update(m4)

    assert len(m4) == len(m5) == 2
    m5.update({10: 11})
    assert len(m5) == 3

    assert repr(m5) == "Map({1: 2, 3: 4, 10: 11})"

    iter(iter(hb.Map({})))


def test_subset(blankfont):
    for planned in (False, True):
        assert blankfont.get_nominal_glyph(ord("a")) == 1
        assert blankfont.get_nominal_glyph(ord("b")) == 2
        assert blankfont.get_nominal_glyph(ord("c")) == 3
        assert blankfont.get_nominal_glyph(ord("d")) == 4
        assert blankfont.get_nominal_glyph(ord("e")) == 5

        inp = hb.SubsetInput()
        inp.sets(hb.SubsetInputSets.UNICODE).set({ord("b")})
        s = inp.sets(hb.SubsetInputSets.LAYOUT_FEATURE_TAG)
        s.clear()
        s.invert()
        inp.layout_script_tag_set.invert()
        inp.unicode_set.update(ord(c) for c in "cd")
        inp.unicode_set.add(ord("e"))

        if not planned:
            face = hb.subset(blankfont.face, inp)
        else:
            plan = hb.SubsetPlan(blankfont.face, inp)
            face = plan.execute()

        assert face is not None
        font = hb.Font(face)

        assert font.get_nominal_glyph(ord("a")) is None
        assert font.get_nominal_glyph(ord("b")) == 1
        assert font.get_nominal_glyph(ord("c")) == 2
        assert font.get_nominal_glyph(ord("d")) == 3
        assert font.get_nominal_glyph(ord("e")) == 4

        blob = face.blob
        assert blob
        assert len(blob) > 100
        face = hb.Face(blob)
        font = hb.Font(face)

        assert font.get_nominal_glyph(ord("a")) is None
        assert font.get_nominal_glyph(ord("b")) == 1
        assert font.get_nominal_glyph(ord("c")) == 2
        assert font.get_nominal_glyph(ord("d")) == 3
        assert font.get_nominal_glyph(ord("e")) == 4

        if planned:
            mapping = plan.old_to_new_glyph_mapping
            reverse = plan.new_to_old_glyph_mapping
            assert 1 not in mapping
            assert mapping[2] == 1
            assert mapping[3] == 2
            assert reverse[mapping[2]] == 2
            assert reverse[mapping[3]] == 3
            assert len(reverse) == 5
            cmap = plan.unicode_to_old_glyph_mapping
            assert cmap[ord("b")] == 2



from libc.stdint cimport uint8_t, uint16_t, uint32_t


cdef extern from "hb.h":

    # hb-common.h
    ctypedef void (*hb_destroy_func_t) (void* user_data)
    ctypedef int hb_bool_t
    ctypedef unsigned long hb_codepoint_t
    ctypedef long hb_position_t
    ctypedef unsigned long hb_mask_t
    ctypedef unsigned long hb_tag_t
    ctypedef uint32_t hb_color_t

    ctypedef enum hb_direction_t:
        HB_DIRECTION_LTR
        HB_DIRECTION_RTL
        HB_DIRECTION_TTB
        HB_DIRECTION_BTT
    ctypedef enum hb_script_t:
        pass
    ctypedef struct hb_language_t:
        pass
    ctypedef struct hb_feature_t:
        hb_tag_t tag
        unsigned long value
        unsigned int start
        unsigned int end
    ctypedef struct hb_glyph_extents_t:
        hb_position_t x_bearing
        hb_position_t y_bearing
        hb_position_t width
        hb_position_t height

    hb_direction_t hb_direction_from_string(const char* str, int len)
    const char* hb_direction_to_string(hb_direction_t direction)
    hb_bool_t hb_feature_from_string(
        const char* str, int len,
        hb_feature_t* feature)
    void hb_feature_to_string(
        hb_feature_t* feature,
        char* buf, unsigned int size)
    hb_language_t hb_language_from_string(const char* str, int len)
    const char* hb_language_to_string(hb_language_t language)
    hb_script_t hb_script_from_string(const char* str, int len)
    hb_tag_t hb_tag_from_string(const char* str, int len)
    void hb_tag_to_string(hb_tag_t tag, char* buf)
    hb_language_t hb_ot_tag_to_language(hb_tag_t tag)
    hb_script_t hb_ot_tag_to_script(hb_tag_t tag)
    const char* hb_version_string()

    uint8_t hb_color_get_alpha(hb_color_t color)
    uint8_t hb_color_get_red(hb_color_t color)
    uint8_t hb_color_get_green(hb_color_t color)
    uint8_t hb_color_get_blue(hb_color_t color)

    hb_color_t HB_COLOR(uint8_t b, uint8_t g, uint8_t r, uint8_t a)

    ctypedef struct hb_user_data_key_t:
        pass

    ctypedef union hb_var_int_t:
        unsigned long u32
        long i32
        unsigned int u16[2]
        int i16[2]
        unsigned short u8[4]
        short i8[4]

    ctypedef union hb_var_num_t:
        float f
        unsigned long u32
        long i32
        unsigned int u16[2]
        int i16[2]
        unsigned short u8[4]
        short i8[4]

    # hb-blob.h
    ctypedef struct hb_blob_t:
        pass
    ctypedef enum hb_memory_mode_t:
        HB_MEMORY_MODE_DUPLICATE
        HB_MEMORY_MODE_READONLY
        HB_MEMORY_MODE_WRITABLE
        HB_MEMORY_MODE_READONLY_MAY_MAKE_WRITABLE

    hb_blob_t* hb_blob_create(
        const char* data, unsigned int length,
        hb_memory_mode_t mode,
        void* user_data, hb_destroy_func_t destroy)

    hb_blob_t* hb_blob_create_from_file(
        const char *file_name)

    hb_blob_t* hb_blob_create_from_file_or_fail(
        const char *file_name)

    hb_blob_t* hb_blob_reference(hb_blob_t* blob)
    void hb_blob_destroy(hb_blob_t* blob)

    const char* hb_blob_get_data(
        hb_blob_t *blob, unsigned int *length)

    unsigned int hb_blob_get_length(
        hb_blob_t *blob)

    hb_blob_t* hb_blob_get_empty()

    # hb-buffer.h
    ctypedef struct hb_buffer_t:
        pass

    cdef hb_codepoint_t HB_BUFFER_REPLACEMENT_CODEPOINT_DEFAULT

    ctypedef enum hb_glyph_flags_t:
        HB_GLYPH_FLAG_UNSAFE_TO_BREAK
        HB_GLYPH_FLAG_UNSAFE_TO_CONCAT
        HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL
        HB_GLYPH_FLAG_DEFINED

    ctypedef struct hb_glyph_info_t:
        hb_codepoint_t codepoint
        hb_mask_t mask
        unsigned long cluster
    ctypedef struct hb_glyph_position_t:
        hb_position_t x_advance
        hb_position_t y_advance
        hb_position_t x_offset
        hb_position_t y_offset
        hb_var_int_t var

    ctypedef enum hb_buffer_content_type_t:
        HB_BUFFER_CONTENT_TYPE_INVALID
        HB_BUFFER_CONTENT_TYPE_UNICODE
        HB_BUFFER_CONTENT_TYPE_GLYPHS

    ctypedef enum hb_buffer_cluster_level_t:
        HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES
        HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
        HB_BUFFER_CLUSTER_LEVEL_CHARACTERS
        HB_BUFFER_CLUSTER_LEVEL_DEFAULT

    ctypedef enum hb_buffer_flags_t:
        HB_BUFFER_FLAG_DEFAULT
        HB_BUFFER_FLAG_BOT
        HB_BUFFER_FLAG_EOT
        HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES
        HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
        HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE
        HB_BUFFER_FLAG_VERIFY
        HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT
        HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL

        HB_BUFFER_FLAG_DEFINED

    hb_buffer_t* hb_buffer_create()
    hb_bool_t hb_buffer_allocation_successful(hb_buffer_t* buffer)
    void hb_buffer_reset(hb_buffer_t *buffer)
    void hb_buffer_clear_contents(hb_buffer_t *buffer)
    void hb_buffer_add_codepoints(
        hb_buffer_t* buffer,
        const hb_codepoint_t* text, int text_length,
        unsigned int item_offset, int item_length)
    void hb_buffer_add_latin1(
        hb_buffer_t* buffer,
        const uint8_t* text, int text_length,
        unsigned int item_offset, int item_length)
    void hb_buffer_add_utf8(
        hb_buffer_t* buffer,
        const char* text, int text_length,
        unsigned int item_offset, int item_length)
    void hb_buffer_add_utf16(
        hb_buffer_t* buffer,
        const uint16_t* text, int text_length,
        unsigned int item_offset, int item_length)
    void hb_buffer_add_utf32(
        hb_buffer_t* buffer,
        const uint32_t* text, int text_length,
        unsigned int item_offset, int item_length)
    void hb_buffer_guess_segment_properties(hb_buffer_t* buffer)
    hb_direction_t hb_buffer_get_direction(hb_buffer_t* buffer)
    void hb_buffer_set_direction(hb_buffer_t* buffer, hb_direction_t direction)
    unsigned int hb_buffer_get_length(const hb_buffer_t *buffer)
    hb_glyph_info_t* hb_buffer_get_glyph_infos(
        hb_buffer_t* buffer, unsigned int* length)
    hb_glyph_position_t* hb_buffer_get_glyph_positions(
        hb_buffer_t* buffer, unsigned int* length)
    hb_script_t hb_buffer_get_script(hb_buffer_t* buffer)
    void hb_buffer_set_script(hb_buffer_t* buffer, hb_script_t script)
    hb_language_t hb_buffer_get_language(hb_buffer_t* buffer)
    void hb_buffer_set_language(hb_buffer_t* buffer, hb_language_t language)
    void hb_buffer_set_cluster_level(hb_buffer_t *buffer,
        hb_buffer_cluster_level_t cluster_level)
    hb_buffer_cluster_level_t hb_buffer_get_cluster_level(hb_buffer_t *buffer)
    void hb_buffer_destroy(hb_buffer_t* buffer)
    ctypedef hb_bool_t (*hb_buffer_message_func_t) (
        hb_buffer_t *buffer,
        hb_font_t *font,
        const char *message,
        void *user_data)
    void hb_buffer_set_message_func(
        hb_buffer_t *buffer,
        hb_buffer_message_func_t func,
        void *user_data,
        void* destroy)
    void hb_buffer_set_flags(hb_buffer_t *buffer, hb_buffer_flags_t  flags)
    hb_buffer_flags_t hb_buffer_get_flags(const hb_buffer_t *buffer)
    void hb_buffer_set_content_type(hb_buffer_t *buffer, hb_buffer_content_type_t  content_type)
    hb_buffer_content_type_t hb_buffer_get_content_type(const hb_buffer_t *buffer)
    void hb_buffer_set_replacement_codepoint(hb_buffer_t *buffer, hb_codepoint_t  replacement)
    hb_codepoint_t hb_buffer_get_replacement_codepoint(const hb_buffer_t *buffer)
    void hb_buffer_set_invisible_glyph(hb_buffer_t *buffer, hb_codepoint_t  invisible)
    hb_codepoint_t hb_buffer_get_invisible_glyph(const hb_buffer_t *buffer)
    void hb_buffer_set_not_found_glyph(hb_buffer_t *buffer, hb_codepoint_t  not_found)
    hb_codepoint_t hb_buffer_get_not_found_glyph(const hb_buffer_t *buffer)


    # hb-face.h
    ctypedef struct hb_face_t:
        pass
    ctypedef hb_blob_t* (*hb_reference_table_func_t) (
        hb_face_t* face, hb_tag_t tag, void* user_data)

    hb_face_t* hb_face_create(hb_blob_t* blob, unsigned int index)
    hb_face_t* hb_face_create_for_tables(
        hb_reference_table_func_t reference_table_func,
        void* user_data, hb_destroy_func_t destroy)
    unsigned int hb_face_get_index(const hb_face_t* face)
    void hb_face_set_index(hb_face_t* face, unsigned int index)
    unsigned int hb_face_get_upem(hb_face_t* face)
    void hb_face_set_upem(hb_face_t* face, unsigned int upem)
    unsigned int hb_face_get_glyph_count(hb_face_t* face)
    void hb_face_set_glyph_count(hb_face_t* face, unsigned int glyph_count)
    void* hb_face_get_user_data(hb_face_t* face, hb_user_data_key_t* key)
    hb_bool_t hb_face_set_user_data(
        hb_face_t* face,
        hb_user_data_key_t* key,
        void* data, hb_destroy_func_t destroy,
        hb_bool_t replace)
    void hb_face_destroy(hb_face_t* face)
    hb_blob_t* hb_face_reference_blob(hb_face_t *face)
    hb_face_t* hb_face_reference(hb_face_t *face)
    hb_face_t* hb_face_get_empty()
    unsigned int hb_face_get_table_tags(
        const hb_face_t *face,
        unsigned int  start_offset,
        unsigned int *table_count,
        hb_tag_t     *table_tags)
    void hb_face_collect_unicodes(hb_face_t *face, hb_set_t *out)
    void hb_face_collect_variation_selectors(hb_face_t *face, hb_set_t *out)
    void hb_face_collect_variation_unicodes(hb_face_t *face, hb_codepoint_t variation_selector, hb_set_t *out)


    # hb-font.h
    ctypedef struct hb_font_funcs_t:
        pass
    ctypedef struct hb_font_t:
        pass
    ctypedef hb_bool_t (*hb_font_get_nominal_glyph_func_t) (
        hb_font_t* font, void* font_data,
        hb_codepoint_t unicode,
        hb_codepoint_t* glyph,
        void* user_data)
    ctypedef hb_bool_t (*hb_font_get_variation_glyph_func_t) (
        hb_font_t *font, void *font_data,
        hb_codepoint_t unicode,
        hb_codepoint_t variation_selector,
        hb_codepoint_t *glyph,
        void *user_data)
    ctypedef hb_position_t (*hb_font_get_glyph_advance_func_t) (
        hb_font_t* font, void* font_data,
        hb_codepoint_t glyph,
        void* user_data)
    ctypedef hb_font_get_glyph_advance_func_t hb_font_get_glyph_h_advance_func_t
    ctypedef hb_font_get_glyph_advance_func_t hb_font_get_glyph_v_advance_func_t
    ctypedef hb_bool_t (*hb_font_get_glyph_origin_func_t) (
        hb_font_t* font, void* font_data,
        hb_codepoint_t glyph,
        hb_position_t* x, hb_position_t* y,
        void* user_data)
    ctypedef hb_font_get_glyph_origin_func_t hb_font_get_glyph_v_origin_func_t
    ctypedef hb_bool_t (*hb_font_get_glyph_name_func_t) (
        hb_font_t *font, void *font_data,
        hb_codepoint_t glyph,
        char *name, unsigned int size,
        void *user_data)
    ctypedef hb_bool_t (*hb_font_get_font_extents_func_t) (
        hb_font_t *font, void *font_data,
        hb_font_extents_t *extents,
        void *user_data)
    ctypedef hb_font_get_font_extents_func_t hb_font_get_font_h_extents_func_t
    ctypedef hb_font_get_font_extents_func_t hb_font_get_font_v_extents_func_t
    ctypedef struct hb_variation_t:
        hb_tag_t tag
        float value
    ctypedef struct hb_font_extents_t:
        hb_position_t ascender
        hb_position_t descender
        hb_position_t line_gap
        hb_position_t reserved9
        hb_position_t reserved8
        hb_position_t reserved7
        hb_position_t reserved6
        hb_position_t reserved5
        hb_position_t reserved4
        hb_position_t reserved3
        hb_position_t reserved2
        hb_position_t reserved1

    hb_font_t* hb_font_create(hb_face_t* face)
    hb_font_t* hb_font_create_sub_font(hb_font_t* parent)
    hb_font_t* hb_font_get_empty()
    hb_font_t* hb_font_reference(hb_font_t *font)
    void hb_font_destroy(hb_font_t* font)

    hb_face_t *hb_font_get_face(hb_font_t *font)
    hb_font_funcs_t* hb_font_funcs_create()
    void hb_font_funcs_set_glyph_h_advance_func(
        hb_font_funcs_t* ffuncs,
        hb_font_get_glyph_h_advance_func_t func,
        void* user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_glyph_v_advance_func(
        hb_font_funcs_t* ffuncs,
        hb_font_get_glyph_v_advance_func_t func,
        void* user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_glyph_v_origin_func(
        hb_font_funcs_t* ffuncs,
        hb_font_get_glyph_v_origin_func_t func,
        void* user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_glyph_name_func(
        hb_font_funcs_t* ffuncs,
        hb_font_get_glyph_name_func_t func,
        void* user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_nominal_glyph_func(
        hb_font_funcs_t* ffuncs,
        hb_font_get_nominal_glyph_func_t func,
        void* user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_variation_glyph_func(
        hb_font_funcs_t *ffuncs,
        hb_font_get_variation_glyph_func_t func,
        void *user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_font_h_extents_func(
        hb_font_funcs_t *ffuncs,
        hb_font_get_font_h_extents_func_t func,
        void *user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_set_font_v_extents_func(
        hb_font_funcs_t *ffuncs,
        hb_font_get_font_v_extents_func_t func,
        void *user_data, hb_destroy_func_t destroy)
    void hb_font_funcs_destroy(hb_font_funcs_t* ffuncs)

    void hb_font_set_funcs(
        hb_font_t* font,
        hb_font_funcs_t* klass,
        void* font_data, hb_destroy_func_t destroy)
    void hb_font_get_scale(hb_font_t* font, int* x_scale, int* y_scale)
    void hb_font_set_scale(hb_font_t* font, int x_scale, int y_scale)
    void hb_font_get_ppem(hb_font_t* font, unsigned int* x_ppem, unsigned int* y_ppem)
    void hb_font_set_ppem(hb_font_t* font, unsigned int x_ppem, unsigned int y_ppem)
    float hb_font_get_ptem(hb_font_t* font)
    void hb_font_set_ptem(hb_font_t* font, float ptem)
    void hb_font_get_synthetic_bold(hb_font_t *font,
                            float *x_embolden,
                            float *y_embolden,
                            hb_bool_t *in_place)
    void hb_font_set_synthetic_bold(hb_font_t *font,
                            float x_embolden,
                            float y_embolden,
                            hb_bool_t in_place)
    float hb_font_get_synthetic_slant(hb_font_t *font)
    void hb_font_set_synthetic_slant(hb_font_t *font, float slant)
    void hb_font_set_variations(
        hb_font_t* font,
        const hb_variation_t* variations,
        unsigned int variations_length)
    void hb_font_set_variation(
        hb_font_t *font,
        hb_tag_t tag,
        float value)
    void hb_font_set_var_named_instance(hb_font_t *font, unsigned int instance_index)
    unsigned int hb_font_get_var_named_instance(hb_font_t *font)
    void hb_font_set_var_coords_design(
        hb_font_t *font,
        const float *coords,
        unsigned int coords_length)
    const float * hb_font_get_var_coords_design(
        hb_font_t *font, unsigned int *length)
    hb_bool_t hb_font_get_glyph_name(
        hb_font_t* font,
        hb_codepoint_t glyph,
        char* name,
        unsigned int size)
    hb_bool_t hb_font_get_glyph_from_name(
        hb_font_t *font,
        const char *name,
        int len,
        hb_codepoint_t *glyph)
    hb_bool_t hb_font_get_glyph_extents(
        hb_font_t* font,
        hb_codepoint_t glyph,
        hb_glyph_extents_t *extents)
    void hb_font_get_extents_for_direction(hb_font_t *font,
        hb_direction_t direction, hb_font_extents_t *extents)
    hb_bool_t hb_font_get_h_extents(hb_font_t *font, hb_font_extents_t *extents)
    hb_bool_t hb_font_get_v_extents(hb_font_t *font, hb_font_extents_t *extents)
    hb_position_t hb_font_get_glyph_h_advance(
        hb_font_t *font,
        hb_codepoint_t glyph)
    hb_position_t hb_font_get_glyph_v_advance(
        hb_font_t *font,
        hb_codepoint_t glyph)
    hb_bool_t hb_font_get_glyph_h_origin(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_position_t *x,
        hb_position_t *y)
    hb_bool_t hb_font_get_glyph_v_origin(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_position_t *x,
        hb_position_t *y)
    hb_bool_t hb_font_get_nominal_glyph(
        hb_font_t *font,
        hb_codepoint_t unicode,
        hb_codepoint_t *glyph)
    hb_bool_t hb_font_get_variation_glyph(
        hb_font_t *font,
        hb_codepoint_t unicode,
        hb_codepoint_t variation_selector,
        hb_codepoint_t *glyph)
    const int * hb_font_get_var_coords_normalized(
        hb_font_t *font,
        unsigned int *length)
    void hb_font_set_var_coords_normalized(
        hb_font_t *font,
        const int *coords,
        unsigned int coords_length)
    void hb_font_glyph_to_string(
        hb_font_t* font,
        hb_codepoint_t glyph,
        char* name,
        unsigned int size)
    hb_bool_t hb_font_glyph_from_string(
        hb_font_t *font,
        const char *s,
        int len,
        hb_codepoint_t *glyph)

    void hb_font_draw_glyph(
        hb_font_t *font,
        hb_codepoint_t glyph,
        const hb_draw_funcs_t *dfuncs,
        void *draw_data)

    void hb_font_paint_glyph(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_paint_funcs_t *pfuncs,
        void *paint_data,
        unsigned int palette_index,
        hb_color_t foreground)

    # hb-draw.h
    ctypedef struct hb_draw_state_t:
       hb_bool_t path_open
       float path_start_x
       float path_start_y
       float current_x
       float current_y
       hb_var_num_t reserved1
       hb_var_num_t reserved2
       hb_var_num_t reserved3
       hb_var_num_t reserved4
       hb_var_num_t reserved5
       hb_var_num_t reserved6
       hb_var_num_t reserved7

    ctypedef struct hb_draw_funcs_t:
        pass

    ctypedef void (*hb_draw_move_to_func_t) (
        hb_draw_funcs_t *dfuncs,
        void *draw_data,
        hb_draw_state_t *st,
        float to_x,
        float to_y,
        void *user_data)

    ctypedef void (*hb_draw_line_to_func_t) (
        hb_draw_funcs_t *dfuncs,
        void *draw_data,
        hb_draw_state_t *st,
        float to_x,
        float to_y,
        void *user_data)

    ctypedef void (*hb_draw_quadratic_to_func_t) (
        hb_draw_funcs_t *dfuncs,
        void *draw_data,
        hb_draw_state_t *st,
        float control_x,
        float control_y,
        float to_x,
        float to_y,
        void *user_data)

    ctypedef void (*hb_draw_cubic_to_func_t) (
        hb_draw_funcs_t *dfuncs,
        void *draw_data,
        hb_draw_state_t *st,
        float control1_x,
        float control1_y,
        float control2_x,
        float control2_y,
        float to_x,
        float to_y,
        void *user_data)

    ctypedef void (*hb_draw_close_path_func_t) (
        hb_draw_funcs_t *dfuncs,
        void *draw_data,
        hb_draw_state_t *st,
        void *user_data)

    void hb_draw_funcs_set_move_to_func(
        hb_draw_funcs_t* dfuncs,
        hb_draw_move_to_func_t func,
        void *user_data,
        hb_destroy_func_t destroy)

    void hb_draw_funcs_set_line_to_func(
        hb_draw_funcs_t* dfuncs,
        hb_draw_line_to_func_t func,
        void *user_data,
        hb_destroy_func_t destroy)

    void hb_draw_funcs_set_quadratic_to_func(
        hb_draw_funcs_t* dfuncs,
        hb_draw_quadratic_to_func_t func,
        void *user_data,
        hb_destroy_func_t destroy)

    void hb_draw_funcs_set_cubic_to_func(
        hb_draw_funcs_t* dfuncs,
        hb_draw_cubic_to_func_t func,
        void *user_data,
        hb_destroy_func_t destroy)

    void hb_draw_funcs_set_close_path_func(
        hb_draw_funcs_t* dfuncs,
        hb_draw_close_path_func_t func,
        void *user_data,
        hb_destroy_func_t destroy)

    hb_draw_funcs_t* hb_draw_funcs_create()

    void hb_draw_funcs_destroy(hb_draw_funcs_t* funcs)

    # hb-paint.h
    ctypedef struct hb_paint_funcs_t:
        pass

    ctypedef struct hb_color_stop_t:
        float offset
        hb_bool_t is_foreground
        hb_color_t color

    ctypedef enum hb_paint_extend_t:
        HB_PAINT_EXTEND_PAD
        HB_PAINT_EXTEND_REPEAT
        HB_PAINT_EXTEND_REFLECT

    ctypedef struct hb_color_line_t:
        pass

    ctypedef enum hb_paint_composite_mode_t:
        HB_PAINT_COMPOSITE_MODE_CLEAR
        HB_PAINT_COMPOSITE_MODE_SRC
        HB_PAINT_COMPOSITE_MODE_DEST
        HB_PAINT_COMPOSITE_MODE_SRC_OVER
        HB_PAINT_COMPOSITE_MODE_DEST_OVER
        HB_PAINT_COMPOSITE_MODE_SRC_IN
        HB_PAINT_COMPOSITE_MODE_DEST_IN
        HB_PAINT_COMPOSITE_MODE_SRC_OUT
        HB_PAINT_COMPOSITE_MODE_DEST_OUT
        HB_PAINT_COMPOSITE_MODE_SRC_ATOP
        HB_PAINT_COMPOSITE_MODE_DEST_ATOP
        HB_PAINT_COMPOSITE_MODE_XOR
        HB_PAINT_COMPOSITE_MODE_PLUS
        HB_PAINT_COMPOSITE_MODE_SCREEN
        HB_PAINT_COMPOSITE_MODE_OVERLAY
        HB_PAINT_COMPOSITE_MODE_DARKEN
        HB_PAINT_COMPOSITE_MODE_LIGHTEN
        HB_PAINT_COMPOSITE_MODE_COLOR_DODGE
        HB_PAINT_COMPOSITE_MODE_COLOR_BURN
        HB_PAINT_COMPOSITE_MODE_HARD_LIGHT
        HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT
        HB_PAINT_COMPOSITE_MODE_DIFFERENCE
        HB_PAINT_COMPOSITE_MODE_EXCLUSION
        HB_PAINT_COMPOSITE_MODE_MULTIPLY
        HB_PAINT_COMPOSITE_MODE_HSL_HUE
        HB_PAINT_COMPOSITE_MODE_HSL_SATURATION
        HB_PAINT_COMPOSITE_MODE_HSL_COLOR
        HB_PAINT_COMPOSITE_MODE_HSL_LUMINOSITY

    hb_paint_funcs_t *hb_paint_funcs_create()
    hb_paint_funcs_t *hb_paint_funcs_get_empty()
    hb_paint_funcs_t *hb_paint_funcs_reference(hb_paint_funcs_t *funcs)
    void hb_paint_funcs_destroy(hb_paint_funcs_t *funcs)
    hb_bool_t hb_paint_funcs_set_user_data(
        hb_paint_funcs_t *funcs,
        hb_user_data_key_t *key,
        void *data,
        hb_destroy_func_t destroy,
        hb_bool_t replace)
    void *hb_paint_funcs_get_user_data(
        const hb_paint_funcs_t *funcs,
        hb_user_data_key_t *key)
    void hb_paint_funcs_make_immutable(hb_paint_funcs_t *funcs)
    hb_bool_t hb_paint_funcs_is_immutable(hb_paint_funcs_t *funcs)

    unsigned int hb_color_line_get_color_stops(
        hb_color_line_t *color_line,
        unsigned int start,
        unsigned int *count,
        hb_color_stop_t *color_stops)

    hb_paint_extend_t hb_color_line_get_extend(hb_color_line_t *color_line)

    ctypedef void (*hb_paint_push_transform_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        float xx, float yx,
        float xy, float yy,
        float dx, float dy,
        void *user_data)
    ctypedef void (*hb_paint_pop_transform_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        void *user_data)
    ctypedef hb_bool_t (*hb_paint_color_glyph_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_codepoint_t glyph,
        hb_font_t *font,
        void *user_data)
    ctypedef void (*hb_paint_push_clip_glyph_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_codepoint_t glyph,
        hb_font_t *font,
        void *user_data)
    ctypedef void (*hb_paint_push_clip_rectangle_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        float xmin, float ymin,
        float xmax, float ymax,
        void *user_data)
    ctypedef void (*hb_paint_pop_clip_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        void *user_data)
    ctypedef void (*hb_paint_color_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_bool_t is_foreground,
        hb_color_t color,
        void *user_data)
    ctypedef hb_bool_t (*hb_paint_image_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_blob_t *image,
        unsigned int width,
        unsigned int height,
        hb_tag_t format,
        float slant,
        hb_glyph_extents_t *extents,
        void *user_data)
    ctypedef unsigned int (*hb_color_line_get_color_stops_func_t) (
        hb_color_line_t *color_line,
        void *color_line_data,
        unsigned int start,
        unsigned int *count,
        hb_color_stop_t *color_stops,
        void *user_data)
    ctypedef hb_paint_extend_t (*hb_color_line_get_extend_func_t) (
        hb_color_line_t *color_line,
        void *color_line_data,
        void *user_data)
    ctypedef struct hb_color_line_t:
        void *data
        hb_color_line_get_color_stops_func_t get_color_stops
        void *get_color_stops_user_data
        hb_color_line_get_extend_func_t get_extend
        void *get_extend_user_data
        void *reserved0
        void *reserved1
        void *reserved2
        void *reserved3
        void *reserved5
        void *reserved6
        void *reserved7
        void *reserved8

    ctypedef void (*hb_paint_linear_gradient_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_color_line_t *color_line,
        float x0, float y0,
        float x1, float y1,
        float x2, float y2,
        void *user_data)
    ctypedef void (*hb_paint_radial_gradient_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_color_line_t *color_line,
        float x0, float y0, float r0,
        float x1, float y1, float r1,
        void *user_data)
    ctypedef void (*hb_paint_sweep_gradient_func_t)  (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_color_line_t *color_line,
        float x0, float y0,
        float start_angle,
        float end_angle,
        void *user_data)
    ctypedef void (*hb_paint_push_group_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        void *user_data)
    ctypedef void (*hb_paint_pop_group_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        hb_paint_composite_mode_t mode,
        void *user_data)
    ctypedef hb_bool_t (*hb_paint_custom_palette_color_func_t) (
        hb_paint_funcs_t *funcs,
        void *paint_data,
        unsigned int color_index,
        hb_color_t *color,
        void *user_data)

    void hb_paint_funcs_set_push_transform_func(
        hb_paint_funcs_t               *funcs,
        hb_paint_push_transform_func_t  func,
        void                           *user_data,
        hb_destroy_func_t               destroy)
    void hb_paint_funcs_set_pop_transform_func(
        hb_paint_funcs_t              *funcs,
        hb_paint_pop_transform_func_t  func,
        void                          *user_data,
        hb_destroy_func_t              destroy)
    void hb_paint_funcs_set_color_glyph_func(
        hb_paint_funcs_t                *funcs,
        hb_paint_color_glyph_func_t     func,
        void                            *user_data,
        hb_destroy_func_t                destroy)
    void hb_paint_funcs_set_push_clip_glyph_func(
        hb_paint_funcs_t                *funcs,
        hb_paint_push_clip_glyph_func_t  func,
        void                            *user_data,
        hb_destroy_func_t                destroy)
    void hb_paint_funcs_set_push_clip_rectangle_func(
        hb_paint_funcs_t                    *funcs,
        hb_paint_push_clip_rectangle_func_t  func,
        void                                *user_data,
        hb_destroy_func_t                    destroy)
    void hb_paint_funcs_set_pop_clip_func(
        hb_paint_funcs_t         *funcs,
        hb_paint_pop_clip_func_t  func,
        void                     *user_data,
        hb_destroy_func_t         destroy)
    void hb_paint_funcs_set_color_func(
        hb_paint_funcs_t      *funcs,
        hb_paint_color_func_t  func,
        void                  *user_data,
        hb_destroy_func_t      destroy)
    void hb_paint_funcs_set_image_func(
        hb_paint_funcs_t      *funcs,
        hb_paint_image_func_t  func,
        void                  *user_data,
        hb_destroy_func_t      destroy)
    void hb_paint_funcs_set_linear_gradient_func (
        hb_paint_funcs_t                *funcs,
        hb_paint_linear_gradient_func_t  func,
        void                            *user_data,
        hb_destroy_func_t                destroy)
    void hb_paint_funcs_set_radial_gradient_func(
        hb_paint_funcs_t                *funcs,
        hb_paint_radial_gradient_func_t  func,
        void                            *user_data,
        hb_destroy_func_t                destroy)
    void hb_paint_funcs_set_sweep_gradient_func(
        hb_paint_funcs_t               *funcs,
        hb_paint_sweep_gradient_func_t  func,
        void                           *user_data,
        hb_destroy_func_t               destroy)
    void hb_paint_funcs_set_push_group_func(
        hb_paint_funcs_t           *funcs,
        hb_paint_push_group_func_t  func,
        void                       *user_data,
        hb_destroy_func_t           destroy)
    void hb_paint_funcs_set_pop_group_func(
        hb_paint_funcs_t          *funcs,
        hb_paint_pop_group_func_t  func,
        void                       *user_data,
        hb_destroy_func_t           destroy)
    void hb_paint_funcs_set_custom_palette_color_func(
        hb_paint_funcs_t                     *funcs,
        hb_paint_custom_palette_color_func_t  func,
        void                                 *user_data,
        hb_destroy_func_t                     destroy)

    # hb-shape.h
    void hb_shape(
        hb_font_t* font,
        hb_buffer_t* buffer,
        const hb_feature_t* features, unsigned int num_features)

    hb_bool_t hb_shape_full(
        hb_font_t *font,
        hb_buffer_t *buffer,
        const hb_feature_t *features,
        unsigned int num_features,
        char ** shaper_list)


    # hb-map.h
    ctypedef struct hb_map_t:
        pass
    cdef hb_codepoint_t HB_MAP_VALUE_INVALID
    hb_map_t* hb_map_create()
    hb_map_t* hb_map_get_empty()
    hb_map_t* hb_map_reference(hb_map_t* map)
    void hb_map_destroy(hb_map_t* map)
    hb_bool_t hb_map_set_user_data(hb_map_t* map, hb_user_data_key_t* key, void* data, hb_destroy_func_t destroy, hb_bool_t replace)
    void* hb_map_get_user_data(const hb_map_t* map, hb_user_data_key_t* key)
    hb_bool_t hb_map_allocation_successful(const hb_map_t* map)
    hb_map_t* hb_map_copy(const hb_map_t* map)
    void hb_map_clear(hb_map_t* map)
    hb_bool_t hb_map_is_empty(const hb_map_t* map)
    unsigned int hb_map_get_population(const hb_map_t* map)
    hb_bool_t hb_map_is_equal(const hb_map_t* map, const hb_map_t* other)
    unsigned int hb_map_hash(const hb_map_t* map)
    void hb_map_set(hb_map_t* map, hb_codepoint_t key, hb_codepoint_t value)
    hb_codepoint_t hb_map_get(const hb_map_t* map, hb_codepoint_t key)
    void hb_map_del(hb_map_t* map, hb_codepoint_t key)
    hb_bool_t hb_map_has(const hb_map_t* map, hb_codepoint_t key)
    void hb_map_update(hb_map_t* map, const hb_map_t* other)
    hb_bool_t hb_map_next(const hb_map_t* map, int *idx, hb_codepoint_t *key, hb_codepoint_t *value)

    # hb-set.h
    ctypedef struct hb_set_t:
        pass
    cdef hb_codepoint_t HB_SET_VALUE_INVALID
    hb_set_t* hb_set_create()
    hb_set_t* hb_set_get_empty()
    hb_set_t* hb_set_reference(hb_set_t* set)
    void hb_set_destroy(hb_set_t* set)
    hb_bool_t hb_set_set_user_data(hb_set_t* set, hb_user_data_key_t* key, void* data, hb_destroy_func_t destroy, hb_bool_t replace)
    void* hb_set_get_user_data(const hb_set_t* set, hb_user_data_key_t* key)
    hb_bool_t hb_set_allocation_successful(const hb_set_t* set)
    hb_set_t* hb_set_copy(const hb_set_t* set)
    void hb_set_clear(hb_set_t* set)
    hb_bool_t hb_set_is_empty(const hb_set_t* set)
    void hb_set_invert(hb_set_t* set)
    hb_bool_t hb_set_is_inverted(const hb_set_t* set)
    hb_bool_t hb_set_has(const hb_set_t* set, hb_codepoint_t codepoint)
    void hb_set_add(hb_set_t* set, hb_codepoint_t codepoint)
    void hb_set_add_range(hb_set_t* set, hb_codepoint_t first, hb_codepoint_t last)
    void hb_set_add_sorted_array(hb_set_t* set, const hb_codepoint_t* sorted_codepoints, unsigned int num_codepoints)
    void hb_set_del(hb_set_t* set, hb_codepoint_t codepoint)
    void hb_set_del_range(hb_set_t* set, hb_codepoint_t first, hb_codepoint_t last)
    hb_bool_t hb_set_is_equal(const hb_set_t* set, const hb_set_t* other)
    unsigned int hb_set_hash(const hb_set_t* set)
    hb_bool_t hb_set_is_subset(const hb_set_t* set, const hb_set_t* larger_set)
    void hb_set_set(hb_set_t* set, const hb_set_t* other)
    void hb_set_union(hb_set_t* set, const hb_set_t* other)
    void hb_set_intersect(hb_set_t* set, const hb_set_t* other)
    void hb_set_subtract(hb_set_t* set, const hb_set_t* other)
    void hb_set_symmetric_difference(hb_set_t* set, const hb_set_t* other)
    unsigned int hb_set_get_population(const hb_set_t* set)
    hb_codepoint_t hb_set_get_min(const hb_set_t* set)
    hb_codepoint_t hb_set_get_max(const hb_set_t* set)
    hb_bool_t hb_set_next(const hb_set_t* set, hb_codepoint_t* codepoint)
    hb_bool_t hb_set_previous(const hb_set_t* set, hb_codepoint_t* codepoint)
    hb_bool_t hb_set_next_range(const hb_set_t* set, hb_codepoint_t* first, hb_codepoint_t* last)
    hb_bool_t hb_set_previous_range(const hb_set_t* set, hb_codepoint_t* first, hb_codepoint_t* last)
    unsigned int hb_set_next_many(const hb_set_t* set, hb_codepoint_t codepoint, hb_codepoint_t* out, unsigned int size)


cdef extern from "hb-ot.h":
    # hb-ot-layout.h
    unsigned int hb_ot_layout_lookup_get_glyph_alternates(
        hb_face_t* face,
        unsigned lookup_index,
        hb_codepoint_t glyph,
        unsigned start_offset,
        unsigned * alternate_count,
        hb_codepoint_t *alternate_glyphs)
    unsigned int hb_ot_layout_language_get_feature_tags(
        hb_face_t* face,
        hb_tag_t table_tag,
        unsigned int script_index,
        unsigned int language_index,
        unsigned int start_offset,
        unsigned int* feature_count,  # in/out
        hb_tag_t* feature_tags)  # out
    unsigned int hb_ot_layout_script_get_language_tags(
        hb_face_t* face,
        hb_tag_t table_tag,
        unsigned int script_index,
        unsigned int start_offset,
        unsigned int* language_count,  # in/out
        hb_tag_t* language_tags)  # out
    unsigned int hb_ot_layout_table_get_script_tags(
        hb_face_t* face,
        hb_tag_t table_tag,
        unsigned int start_offset,
        unsigned int* script_count,  # in/out
        hb_tag_t* script_tags)  # out

    ctypedef enum hb_ot_layout_baseline_tag_t:
        HB_OT_LAYOUT_BASELINE_TAG_ROMAN
        HB_OT_LAYOUT_BASELINE_TAG_HANGING
        HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
        HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
        HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
        HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
        HB_OT_LAYOUT_BASELINE_TAG_MATH

    hb_bool_t hb_ot_layout_get_baseline(
        hb_font_t* font,
        hb_ot_layout_baseline_tag_t baseline_tag,
        hb_direction_t direction,
        hb_tag_t script_tag,
        hb_tag_t language_tag,
        hb_position_t* coord)  # out

    hb_bool_t hb_ot_layout_has_glyph_classes(hb_face_t *face)
    hb_bool_t hb_ot_layout_has_positioning(hb_face_t *face)
    hb_bool_t hb_ot_layout_has_substitution(hb_face_t *face)

    ctypedef enum hb_ot_layout_glyph_class_t:
        HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED
        HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
        HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE
        HB_OT_LAYOUT_GLYPH_CLASS_MARK
        HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
    hb_ot_layout_glyph_class_t hb_ot_layout_get_glyph_class(hb_face_t *face, hb_codepoint_t glyph)

    # hb-ot-font.h
    void hb_ot_font_set_funcs(hb_font_t* font)

    # hb-ot-name.h
    ctypedef unsigned int hb_ot_name_id_t
    ctypedef enum hb_ot_name_id_predefined_t:
        HB_OT_NAME_ID_COPYRIGHT
        HB_OT_NAME_ID_FONT_FAMILY
        HB_OT_NAME_ID_FONT_SUBFAMILY
        HB_OT_NAME_ID_UNIQUE_ID
        HB_OT_NAME_ID_FULL_NAME
        HB_OT_NAME_ID_VERSION_STRING
        HB_OT_NAME_ID_POSTSCRIPT_NAME
        HB_OT_NAME_ID_TRADEMARK
        HB_OT_NAME_ID_MANUFACTURER
        HB_OT_NAME_ID_DESIGNER
        HB_OT_NAME_ID_DESCRIPTION
        HB_OT_NAME_ID_VENDOR_URL
        HB_OT_NAME_ID_DESIGNER_URL
        HB_OT_NAME_ID_LICENSE
        HB_OT_NAME_ID_LICENSE_URL
        HB_OT_NAME_ID_TYPOGRAPHIC_FAMILY
        HB_OT_NAME_ID_TYPOGRAPHIC_SUBFAMILY
        HB_OT_NAME_ID_MAC_FULL_NAME
        HB_OT_NAME_ID_SAMPLE_TEXT
        HB_OT_NAME_ID_CID_FINDFONT_NAME
        HB_OT_NAME_ID_WWS_FAMILY
        HB_OT_NAME_ID_WWS_SUBFAMILY
        HB_OT_NAME_ID_LIGHT_BACKGROUND
        HB_OT_NAME_ID_DARK_BACKGROUND
        HB_OT_NAME_ID_VARIATIONS_PS_PREFIX
        HB_OT_NAME_ID_INVALID

    # hb-ot-color.h
    hb_bool_t hb_ot_color_has_palettes(hb_face_t *face)
    unsigned int hb_ot_color_palette_get_count(hb_face_t *face)
    hb_ot_name_id_t hb_ot_color_palette_get_name_id(hb_face_t *face, unsigned int palette_index)
    hb_ot_name_id_t hb_ot_color_palette_color_get_name_id(hb_face_t *face, unsigned int color_index)
    ctypedef enum hb_ot_color_palette_flags_t:
        HB_OT_COLOR_PALETTE_FLAG_DEFAULT
        HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND
        HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
    hb_ot_color_palette_flags_t hb_ot_color_palette_get_flags(hb_face_t *face, unsigned int palette_index)
    unsigned int hb_ot_color_palette_get_colors(
        hb_face_t *face,
        unsigned int palette_index,
        unsigned int start_offset,
        unsigned int *color_count,  # IN/OUT. May be NULL.
        hb_color_t *colors)  # OUT. May be NULL.

    hb_bool_t hb_ot_color_has_layers(hb_face_t *face)
    ctypedef struct hb_ot_color_layer_t:
        hb_codepoint_t glyph
        unsigned int color_index
    unsigned int hb_ot_color_glyph_get_layers(
        hb_face_t *face,
        hb_codepoint_t glyph,
        unsigned int start_offset,
        unsigned int *layer_count,  # IN/OUT. May be NULL.
        hb_ot_color_layer_t *layers)  # OUT. May be NULL.

    hb_bool_t hb_ot_color_has_paint(hb_face_t *face)
    hb_bool_t hb_ot_color_glyph_has_paint(hb_face_t *face, hb_codepoint_t glyph)

    hb_bool_t hb_ot_color_has_svg(hb_face_t *face)
    hb_blob_t *hb_ot_color_glyph_reference_svg(hb_face_t *face, hb_codepoint_t glyph)

    hb_bool_t hb_ot_color_has_png(hb_face_t *face)
    hb_blob_t *hb_ot_color_glyph_reference_png(hb_font_t *font, hb_codepoint_t glyph)

    # hb-ot-math.h
    ctypedef enum hb_ot_math_constant_t:
        HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN
        HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
        HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT
        HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT
        HB_OT_MATH_CONSTANT_MATH_LEADING
        HB_OT_MATH_CONSTANT_AXIS_HEIGHT
        HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT
        HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT
        HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX
        HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN
        HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP
        HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED
        HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN
        HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX
        HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN
        HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT
        HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT
        HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN
        HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN
        HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN
        HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN
        HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP
        HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
        HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_STACK_GAP_MIN
        HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN
        HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP
        HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN
        HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN
        HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP
        HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP
        HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN
        HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN
        HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN
        HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS
        HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN
        HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN
        HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP
        HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP
        HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP
        HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS
        HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER
        HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP
        HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS
        HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER
        HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP
        HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP
        HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS
        HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER
        HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE
        HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE
        HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT

    ctypedef enum hb_ot_math_kern_t:
        HB_OT_MATH_KERN_TOP_RIGHT
        HB_OT_MATH_KERN_TOP_LEFT
        HB_OT_MATH_KERN_BOTTOM_RIGHT
        HB_OT_MATH_KERN_BOTTOM_LEFT

    ctypedef enum hb_ot_math_glyph_part_flags_t:
        HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER


    ctypedef struct hb_ot_math_kern_entry_t:
        hb_position_t max_correction_height
        hb_position_t kern_value

    ctypedef struct hb_ot_math_glyph_variant_t:
        hb_codepoint_t glyph
        hb_position_t advance

    ctypedef struct hb_ot_math_glyph_part_t:
        hb_codepoint_t glyph
        hb_position_t start_connector_length
        hb_position_t end_connector_length
        hb_position_t full_advance
        hb_ot_math_glyph_part_flags_t flags

    hb_bool_t hb_ot_math_has_data(hb_face_t *face)

    hb_position_t hb_ot_math_get_constant(
        hb_font_t *font,
        hb_ot_math_constant_t constant)

    hb_position_t hb_ot_math_get_glyph_italics_correction(
        hb_font_t *font,
        hb_codepoint_t glyph)

    hb_position_t hb_ot_math_get_glyph_top_accent_attachment(
        hb_font_t *font,
        hb_codepoint_t glyph)

    hb_bool_t hb_ot_math_is_glyph_extended_shape(
        hb_face_t *face,
        hb_codepoint_t glyph)

    hb_position_t hb_ot_math_get_glyph_kerning(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_ot_math_kern_t kern,
        hb_position_t correction_height)

    unsigned int hb_ot_math_get_glyph_kernings(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_ot_math_kern_t kern,
        unsigned int start_offset,
        unsigned int *entries_count, # in/out
        hb_ot_math_kern_entry_t *kern_entries) # out

    unsigned int hb_ot_math_get_glyph_variants(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_direction_t direction,
        unsigned int start_offset,
        unsigned int *variants_count, # in/out
        hb_ot_math_glyph_variant_t *variants) # out

    hb_position_t hb_ot_math_get_min_connector_overlap(
        hb_font_t *font,
        hb_direction_t direction)

    unsigned int hb_ot_math_get_glyph_assembly(
        hb_font_t *font,
        hb_codepoint_t glyph,
        hb_direction_t direction,
        unsigned int start_offset,
        unsigned int *parts_count, # in/out
        hb_ot_math_glyph_part_t *parts, # out
        hb_position_t *italics_correction) # out

cdef extern from "hb-subset-repacker.h":
    ctypedef struct hb_link_t:
        unsigned int width
        unsigned int position
        unsigned int objidx
    ctypedef struct hb_object_t:
        char *head
        char *tail
        unsigned int num_real_links
        hb_link_t *real_links
        unsigned int num_virtual_links
        hb_link_t *virtual_links

    hb_blob_t* hb_subset_repack_or_fail(
        hb_tag_t table_tag,
        hb_object_t* hb_objects,
        unsigned int num_hb_objs)

cdef extern from "hb-subset.h":
    ctypedef struct hb_subset_input_t:
        pass
    ctypedef struct hb_subset_plan_t:
        pass
    ctypedef enum hb_subset_flags_t:
        HB_SUBSET_FLAGS_DEFAULT
        HB_SUBSET_FLAGS_NO_HINTING
        HB_SUBSET_FLAGS_RETAIN_GIDS
        HB_SUBSET_FLAGS_DESUBROUTINIZE
        HB_SUBSET_FLAGS_NAME_LEGACY
        HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG
        HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED
        HB_SUBSET_FLAGS_NOTDEF_OUTLINE
        HB_SUBSET_FLAGS_GLYPH_NAMES
        HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES
        HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE
    ctypedef enum hb_subset_sets_t:
        HB_SUBSET_SETS_GLYPH_INDEX
        HB_SUBSET_SETS_UNICODE
        HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG
        HB_SUBSET_SETS_DROP_TABLE_TAG
        HB_SUBSET_SETS_NAME_ID
        HB_SUBSET_SETS_NAME_LANG_ID
        HB_SUBSET_SETS_LAYOUT_FEATURE_TAG
        HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG
    hb_subset_input_t* hb_subset_input_create_or_fail()
    hb_subset_input_t* hb_subset_input_reference(hb_subset_input_t* input)
    void hb_subset_input_destroy(hb_subset_input_t* input)
    hb_bool_t hb_subset_input_set_user_data(hb_subset_input_t* input, hb_user_data_key_t* key, void* data, hb_destroy_func_t destroy, hb_bool_t replace)
    void* hb_subset_input_get_user_data(const hb_subset_input_t* input, hb_user_data_key_t* key)
    void hb_subset_input_keep_everything(hb_subset_input_t* input)
    hb_set_t* hb_subset_input_unicode_set(hb_subset_input_t* input)
    hb_set_t* hb_subset_input_glyph_set(hb_subset_input_t* input)
    hb_set_t* hb_subset_input_set(hb_subset_input_t* input, hb_subset_sets_t set_type)
    hb_subset_flags_t hb_subset_input_get_flags(hb_subset_input_t* input)
    void hb_subset_input_set_flags(hb_subset_input_t* input, unsigned value)
    hb_bool_t hb_subset_input_pin_axis_to_default(hb_subset_input_t* input, hb_face_t* face, hb_tag_t axis_tag)
    hb_bool_t hb_subset_input_pin_axis_location(hb_subset_input_t* input, hb_face_t* face, hb_tag_t axis_tag, float axis_value)
    hb_face_t* hb_subset_preprocess(hb_face_t* source)
    hb_face_t* hb_subset_or_fail(hb_face_t* source, const hb_subset_input_t* input)
    hb_face_t* hb_subset_plan_execute_or_fail(hb_subset_plan_t* plan)
    hb_subset_plan_t* hb_subset_plan_create_or_fail(hb_face_t* face, const hb_subset_input_t* input)
    void hb_subset_plan_destroy(hb_subset_plan_t* plan)
    const hb_map_t* hb_subset_plan_old_to_new_glyph_mapping(const hb_subset_plan_t* plan)
    const hb_map_t* hb_subset_plan_new_to_old_glyph_mapping(const hb_subset_plan_t* plan)
    const hb_map_t* hb_subset_plan_unicode_to_old_glyph_mapping(const hb_subset_plan_t* plan)
    hb_subset_plan_t* hb_subset_plan_reference(hb_subset_plan_t* plan)
    hb_bool_t hb_subset_plan_set_user_data(hb_subset_plan_t* plan, hb_user_data_key_t* key, void* data, hb_destroy_func_t destroy, hb_bool_t replace)
    void* hb_subset_plan_get_user_data(const hb_subset_plan_t* plan, hb_user_data_key_t* key)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import platform
from io import open
from typing import List

import pkgconfig
from Cython.Build import cythonize
from setuptools import Extension, setup

def bool_from_environ(key: str):
    value = os.environ.get(key)
    if not value:
        return False
    if value == "1":
        return True
    if value == "0":
        return False
    raise ValueError(f"Environment variable {key} has invalid value {value}. Please set it to 1, 0 or an empty string")

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

use_system_libraries = bool_from_environ("USE_SYSTEM_LIBS")
use_cython_linetrace = bool_from_environ("CYTHON_LINETRACE")
use_cython_annotate = bool_from_environ("CYTHON_ANNOTATE")

def _configure_extensions_with_system_libs() -> List[Extension]:
    include_dirs = []
    define_macros = []
    libraries = []
    library_dirs = []

    # There is no way (or, at least, no reliable one) to check whether `harfbuzz` has been built with experimental API support.
    #
    # meson-generated CMake[1] and pkg-config[2] files don't include harfbuzz define macros, neither separately nor as `-D` defines in `cflags`.
    #
    # `config.h`[3] and `hb-config.hh`[4] are generated at buildtime but treated as internal headers. That is, they are only used to compile harfbuzz,
    # but are never installed nor exposed to the library consumers in other ways. Some configuration parameters (e.g., either harfbuzz has been built
    # with `directwrite` support) are exposed via `hb-features.h`[5, 6] header, but not the one associated with experimental API.
    #
    # One solution here could be, patch `harfbuzz` to expose so, but the quickest and "cheapest" fix is just force-turn the experimental API in headers.
    # If native `harfbuzz` is built without experimental API support, `uharfbuzz` will fail to build with undefined symbols errors.
    #
    # References:
    # 1. https://github.com/harfbuzz/harfbuzz/blob/9.0.0/src/harfbuzz-config.cmake.in
    # 2. https://github.com/harfbuzz/harfbuzz/blob/9.0.0/src/harfbuzz.pc.in
    # 3. https://github.com/harfbuzz/harfbuzz/blob/9.0.0/meson.build#L461
    # 4. https://github.com/harfbuzz/harfbuzz/blob/main/src/hb-config.hh
    # 5. https://github.com/harfbuzz/harfbuzz/blob/9.0.0/src/hb-features.h.in
    # 6. https://github.com/harfbuzz/harfbuzz/pull/3847
    define_macros.append(("HB_EXPERIMENTAL_API", "1"))

    harfbuzz_components = ["harfbuzz-subset"]
    for harfbuzz_component in harfbuzz_components:
        harfbuzz_component_configuration = pkgconfig.parse(harfbuzz_component)
        include_dirs += harfbuzz_component_configuration["include_dirs"]
        define_macros += harfbuzz_component_configuration["define_macros"]
        libraries += harfbuzz_component_configuration["libraries"]
        library_dirs += harfbuzz_component_configuration["library_dirs"]

    if use_cython_linetrace:
        define_macros.append(("CYTHON_TRACE_NOGIL", "1"))

    extension = Extension(
        "uharfbuzz._harfbuzz",
        define_macros=define_macros,
        include_dirs=include_dirs,
        sources=[
            "src/uharfbuzz/_harfbuzz.pyx",
        ],
        language="c++",
        libraries=libraries,
        library_dirs=library_dirs
    )

    extension_test = Extension(
        "uharfbuzz._harfbuzz_test",
        define_macros=define_macros,
        include_dirs=include_dirs,
        sources=[
            "src/uharfbuzz/_draw_test_funcs.cc",
            "src/uharfbuzz/_harfbuzz_test.pyx",
        ],
        language="c++",
        libraries=libraries,
        library_dirs=library_dirs
    )

    return [extension, extension_test]

def _configure_extensions_with_vendored_libs() -> List[Extension]:
    define_macros = [("HB_NO_MT", "1"), ("HB_EXPERIMENTAL_API", "1")]
    if use_cython_linetrace:
        define_macros.append(("CYTHON_TRACE_NOGIL", "1"))

    extra_compile_args = []
    extra_link_args = []
    libraries = []
    if platform.system() != "Windows":
        extra_compile_args.append("-std=c++11")
        define_macros.append(("HAVE_MMAP", "1"))
        define_macros.append(("HAVE_UNISTD_H", "1"))
        define_macros.append(("HAVE_SYS_MMAN_H", "1"))
    else:
        define_macros.append(("HAVE_DIRECTWRITE", "1"))
        define_macros.append(("HAVE_UNISCRIBE", "1"))
        libraries += ["usp10", "gdi32", "user32", "rpcrt4", "dwrite"]

    if platform.system() == "Darwin":
        define_macros.append(("HAVE_CORETEXT", "1"))
        extra_link_args.extend(["-framework", "ApplicationServices"])

    extension = Extension(
        "uharfbuzz._harfbuzz",
        define_macros=define_macros,
        include_dirs=["harfbuzz/src"],
        sources=[
            "harfbuzz/src/harfbuzz-subset.cc",
            "harfbuzz/src/hb-coretext.cc",
            "harfbuzz/src/hb-directwrite.cc",
            "harfbuzz/src/hb-uniscribe.cc",
            "src/uharfbuzz/_harfbuzz.pyx",
        ],
        language="c++",
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )

    extension_test = Extension(
        "uharfbuzz._harfbuzz_test",
        define_macros=define_macros,
        include_dirs=["harfbuzz/src"],
        sources=[
            "src/uharfbuzz/_draw_test_funcs.cc",
            "src/uharfbuzz/_harfbuzz_test.pyx",
        ],
        language="c++",
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )

    return [extension, extension_test]

def configure_extensions() -> List[Extension]:
    if use_system_libraries:
        return _configure_extensions_with_system_libs()
    else:
        return _configure_extensions_with_vendored_libs()

setup(
    name="uharfbuzz",
    use_scm_version={"write_to": "src/uharfbuzz/_version.py"},
    description="Streamlined Cython bindings for the harfbuzz shaping engine",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Adrien Tétar",
    author_email="adri-from-59@hotmail.fr",
    url="https://github.com/trufont/uharfbuzz",
    license="Apache License 2.0",
    package_dir={"": "src"},
    packages=["uharfbuzz"],
    zip_safe=False,
    setup_requires=["setuptools_scm"],
    python_requires=">=3.5",
    ext_modules=cythonize(
        configure_extensions(),
        annotate=use_cython_annotate,
        compiler_directives={"linetrace": use_cython_linetrace},
    ),
)
